"""
    pygments.lexers.python
    ~~~~~~~~~~~~~~~~~~~~~~

    Lexers for Python and related languages.

    :copyright: Copyright 2006-2022 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

import re
import keyword

from pip._vendor.pygments.lexer import (
    Lexer,
    RegexLexer,
    include,
    bygroups,
    using,
    default,
    words,
    combined,
    do_insertions,
    this,
)
from pip._vendor.pygments.util import get_bool_opt, shebang_matches
from pip._vendor.pygments.token import (
    Text,
    Comment,
    Operator,
    Keyword,
    Name,
    String,
    Number,
    Punctuation,
    Generic,
    Other,
    Error,
)
from pip._vendor.pygments import unistring as uni

__all__ = [
    "PythonLexer",
    "PythonConsoleLexer",
    "PythonTracebackLexer",
    "Python2Lexer",
    "Python2TracebackLexer",
    "CythonLexer",
    "DgLexer",
    "NumPyLexer",
]

line_re = re.compile(".*?\n")


class PythonLexer(RegexLexer):
    """
    For Python source code (version 3.x).

    .. versionadded:: 0.10

    .. versionchanged:: 2.5
       This is now the default ``PythonLexer``.  It is still available as the
       alias ``Python3Lexer``.
    """

    name = "Python"
    url = "http://www.python.org"
    aliases = ["python", "py", "sage", "python3", "py3"]
    filenames = [
        "*.py",
        "*.pyw",
        # Jython
        "*.jy",
        # Sage
        "*.sage",
        # SCons
        "*.sc",
        "SConstruct",
        "SConscript",
        # Skylark/Starlark (used by Bazel, Buck, and Pants)
        "*.bzl",
        "BUCK",
        "BUILD",
        "BUILD.bazel",
        "WORKSPACE",
        # Twisted Application infrastructure
        "*.tac",
    ]
    mimetypes = [
        "text/x-python",
        "application/x-python",
        "text/x-python3",
        "application/x-python3",
    ]

    uni_name = "[%s][%s]*" % (uni.xid_start, uni.xid_continue)

    def innerstring_rules(ttype):
        return [
            # the old style '%s' % (...) string formatting (still valid in Py3)
            (
                r"%(\(\w+\))?[-#0 +]*([0-9]+|[*])?(\.([0-9]+|[*]))?"
                "[hlL]?[E-GXc-giorsaux%]",
                String.Interpol,
            ),
            # the new style '{}'.format(...) string formatting
            (
                r"\{"
                r"((\w+)((\.\w+)|(\[[^\]]+\]))*)?"  # field name
                r"(\![sra])?"  # conversion
                r"(\:(.?[<>=\^])?[-+ ]?#?0?(\d+)?,?(\.\d+)?[E-GXb-gnosx%]?)?"
                r"\}",
                String.Interpol,
            ),
            # backslashes, quotes and formatting signs must be parsed one at a time
            (r'[^\\\'"%{\n]+', ttype),
            (r'[\'"\\]', ttype),
            # unhandled string formatting sign
            (r"%|(\{{1,2})", ttype)
            # newlines are an error (use "nl" state)
        ]

    def fstring_rules(ttype):
        return [
            # Assuming that a '}' is the closing brace after format specifier.
            # Sadly, this means that we won't detect syntax error. But it's
            # more important to parse correct syntax correctly, than to
            # highlight invalid syntax.
            (r"\}", String.Interpol),
            (r"\{", String.Interpol, "expr-inside-fstring"),
            # backslashes, quotes and formatting signs must be parsed one at a time
            (r'[^\\\'"{}\n]+', ttype),
            (r'[\'"\\]', ttype),
            # newlines are an error (use "nl" state)
        ]

    tokens = {
        "root": [
            (r"\n", Text),
            (
                r'^(\s*)([rRuUbB]{,2})("""(?:.|\n)*?""")',
                bygroups(Text, String.Affix, String.Doc),
            ),
            (
                r"^(\s*)([rRuUbB]{,2})('''(?:.|\n)*?''')",
                bygroups(Text, String.Affix, String.Doc),
            ),
            (r"\A#!.+$", Comment.Hashbang),
            (r"#.*$", Comment.Single),
            (r"\\\n", Text),
            (r"\\", Text),
            include("keywords"),
            include("soft-keywords"),
            (r"(def)((?:\s|\\\s)+)", bygroups(Keyword, Text), "funcname"),
            (r"(class)((?:\s|\\\s)+)", bygroups(Keyword, Text), "classname"),
            (r"(from)((?:\s|\\\s)+)", bygroups(Keyword.Namespace, Text), "fromimport"),
            (r"(import)((?:\s|\\\s)+)", bygroups(Keyword.Namespace, Text), "import"),
            include("expr"),
        ],
        "expr": [
            # raw f-strings
            (
                '(?i)(rf|fr)(""")',
                bygroups(String.Affix, String.Double),
                combined("rfstringescape", "tdqf"),
            ),
            (
                "(?i)(rf|fr)(''')",
                bygroups(String.Affix, String.Single),
                combined("rfstringescape", "tsqf"),
            ),
            (
                '(?i)(rf|fr)(")',
                bygroups(String.Affix, String.Double),
                combined("rfstringescape", "dqf"),
            ),
            (
                "(?i)(rf|fr)(')",
                bygroups(String.Affix, String.Single),
                combined("rfstringescape", "sqf"),
            ),
            # non-raw f-strings
            (
                '([fF])(""")',
                bygroups(String.Affix, String.Double),
                combined("fstringescape", "tdqf"),
            ),
            (
                "([fF])(''')",
                bygroups(String.Affix, String.Single),
                combined("fstringescape", "tsqf"),
            ),
            (
                '([fF])(")',
                bygroups(String.Affix, String.Double),
                combined("fstringescape", "dqf"),
            ),
            (
                "([fF])(')",
                bygroups(String.Affix, String.Single),
                combined("fstringescape", "sqf"),
            ),
            # raw bytes and strings
            ('(?i)(rb|br|r)(""")', bygroups(String.Affix, String.Double), "tdqs"),
            ("(?i)(rb|br|r)(''')", bygroups(String.Affix, String.Single), "tsqs"),
            ('(?i)(rb|br|r)(")', bygroups(String.Affix, String.Double), "dqs"),
            ("(?i)(rb|br|r)(')", bygroups(String.Affix, String.Single), "sqs"),
            # non-raw strings
            (
                '([uU]?)(""")',
                bygroups(String.Affix, String.Double),
                combined("stringescape", "tdqs"),
            ),
            (
                "([uU]?)(''')",
                bygroups(String.Affix, String.Single),
                combined("stringescape", "tsqs"),
            ),
            (
                '([uU]?)(")',
                bygroups(String.Affix, String.Double),
                combined("stringescape", "dqs"),
            ),
            (
                "([uU]?)(')",
                bygroups(String.Affix, String.Single),
                combined("stringescape", "sqs"),
            ),
            # non-raw bytes
            (
                '([bB])(""")',
                bygroups(String.Affix, String.Double),
                combined("bytesescape", "tdqs"),
            ),
            (
                "([bB])(''')",
                bygroups(String.Affix, String.Single),
                combined("bytesescape", "tsqs"),
            ),
            (
                '([bB])(")',
                bygroups(String.Affix, String.Double),
                combined("bytesescape", "dqs"),
            ),
            (
                "([bB])(')",
                bygroups(String.Affix, String.Single),
                combined("bytesescape", "sqs"),
            ),
            (r"[^\S\n]+", Text),
            include("numbers"),
            (r"!=|==|<<|>>|:=|[-~+/*%=<>&^|.]", Operator),
            (r"[]{}:(),;[]", Punctuation),
            (r"(in|is|and|or|not)\b", Operator.Word),
            include("expr-keywords"),
            include("builtins"),
            include("magicfuncs"),
            include("magicvars"),
            include("name"),
        ],
        "expr-inside-fstring": [
            (r"[{([]", Punctuation, "expr-inside-fstring-inner"),
            # without format specifier
            (
                r"(=\s*)?"  # debug (https://bugs.python.org/issue36817)
                r"(\![sraf])?"  # conversion
                r"\}",
                String.Interpol,
                "#pop",
            ),
            # with format specifier
            # we'll catch the remaining '}' in the outer scope
            (
                r"(=\s*)?"  # debug (https://bugs.python.org/issue36817)
                r"(\![sraf])?"  # conversion
                r":",
                String.Interpol,
                "#pop",
            ),
            (r"\s+", Text),  # allow new lines
            include("expr"),
        ],
        "expr-inside-fstring-inner": [
            (r"[{([]", Punctuation, "expr-inside-fstring-inner"),
            (r"[])}]", Punctuation, "#pop"),
            (r"\s+", Text),  # allow new lines
            include("expr"),
        ],
        "expr-keywords": [
            # Based on https://docs.python.org/3/reference/expressions.html
            (
                words(
                    (
                        "async for",
                        "await",
                        "else",
                        "for",
                        "if",
                        "lambda",
                        "yield",
                        "yield from",
                    ),
                    suffix=r"\b",
                ),
                Keyword,
            ),
            (words(("True", "False", "None"), suffix=r"\b"), Keyword.Constant),
        ],
        "keywords": [
            (
                words(
                    (
                        "assert",
                        "async",
                        "await",
                        "break",
                        "continue",
                        "del",
                        "elif",
                        "else",
                        "except",
                        "finally",
                        "for",
                        "global",
                        "if",
                        "lambda",
                        "pass",
                        "raise",
                        "nonlocal",
                        "return",
                        "try",
                        "while",
                        "yield",
                        "yield from",
                        "as",
                        "with",
                    ),
                    suffix=r"\b",
                ),
                Keyword,
            ),
            (words(("True", "False", "None"), suffix=r"\b"), Keyword.Constant),
        ],
        "soft-keywords": [
            # `match`, `case` and `_` soft keywords
            (
                r"(^[ \t]*)"  # at beginning of line + possible indentation
                r"(match|case)\b"  # a possible keyword
                r"(?![ \t]*(?:"  # not followed by...
                r"[:,;=^&|@~)\]}]|(?:"
                + r"|".join(  # characters and keywords that mean this isn't
                    keyword.kwlist
                )
                + r")\b))",  # pattern matching
                bygroups(Text, Keyword),
                "soft-keywords-inner",
            ),
        ],
        "soft-keywords-inner": [
            # optional `_` keyword
            (r"(\s+)([^\n_]*)(_\b)", bygroups(Text, using(this), Keyword)),
            default("#pop"),
        ],
        "builtins": [
            (
                words(
                    (
                        "__import__",
                        "abs",
                        "all",
                        "any",
                        "bin",
                        "bool",
                        "bytearray",
                        "breakpoint",
                        "bytes",
                        "chr",
                        "classmethod",
                        "compile",
                        "complex",
                        "delattr",
                        "dict",
                        "dir",
                        "divmod",
                        "enumerate",
                        "eval",
                        "filter",
                        "float",
                        "format",
                        "frozenset",
                        "getattr",
                        "globals",
                        "hasattr",
                        "hash",
                        "hex",
                        "id",
                        "input",
                        "int",
                        "isinstance",
                        "issubclass",
                        "iter",
                        "len",
                        "list",
                        "locals",
                        "map",
                        "max",
                        "memoryview",
                        "min",
                        "next",
                        "object",
                        "oct",
                        "open",
                        "ord",
                        "pow",
                        "print",
                        "property",
                        "range",
                        "repr",
                        "reversed",
                        "round",
                        "set",
                        "setattr",
                        "slice",
                        "sorted",
                        "staticmethod",
                        "str",
                        "sum",
                        "super",
                        "tuple",
                        "type",
                        "vars",
                        "zip",
                    ),
                    prefix=r"(?<!\.)",
                    suffix=r"\b",
                ),
                Name.Builtin,
            ),
            (r"(?<!\.)(self|Ellipsis|NotImplemented|cls)\b", Name.Builtin.Pseudo),
            (
                words(
                    (
                        "ArithmeticError",
                        "AssertionError",
                        "AttributeError",
                        "BaseException",
                        "BufferError",
                        "BytesWarning",
                        "DeprecationWarning",
                        "EOFError",
                        "EnvironmentError",
                        "Exception",
                        "FloatingPointError",
                        "FutureWarning",
                        "GeneratorExit",
                        "IOError",
                        "ImportError",
                        "ImportWarning",
                        "IndentationError",
                        "IndexError",
                        "KeyError",
                        "KeyboardInterrupt",
                        "LookupError",
                        "MemoryError",
                        "NameError",
                        "NotImplementedError",
                        "OSError",
                        "OverflowError",
                        "PendingDeprecationWarning",
                        "ReferenceError",
                        "ResourceWarning",
                        "RuntimeError",
                        "RuntimeWarning",
                        "StopIteration",
                        "SyntaxError",
                        "SyntaxWarning",
                        "SystemError",
                        "SystemExit",
                        "TabError",
                        "TypeError",
                        "UnboundLocalError",
                        "UnicodeDecodeError",
                        "UnicodeEncodeError",
                        "UnicodeError",
                        "UnicodeTranslateError",
                        "UnicodeWarning",
                        "UserWarning",
                        "ValueError",
                        "VMSError",
                        "Warning",
                        "WindowsError",
                        "ZeroDivisionError",
                        # new builtin exceptions from PEP 3151
                        "BlockingIOError",
                        "ChildProcessError",
                        "ConnectionError",
                        "BrokenPipeError",
                        "ConnectionAbortedError",
                        "ConnectionRefusedError",
                        "ConnectionResetError",
                        "FileExistsError",
                        "FileNotFoundError",
                        "InterruptedError",
                        "IsADirectoryError",
                        "NotADirectoryError",
                        "PermissionError",
                        "ProcessLookupError",
                        "TimeoutError",
                        # others new in Python 3
                        "StopAsyncIteration",
                        "ModuleNotFoundError",
                        "RecursionError",
                        "EncodingWarning",
                    ),
                    prefix=r"(?<!\.)",
                    suffix=r"\b",
                ),
                Name.Exception,
            ),
        ],
        "magicfuncs": [
            (
                words(
                    (
                        "__abs__",
                        "__add__",
                        "__aenter__",
                        "__aexit__",
                        "__aiter__",
                        "__and__",
                        "__anext__",
                        "__await__",
                        "__bool__",
                        "__bytes__",
                        "__call__",
                        "__complex__",
                        "__contains__",
                        "__del__",
                        "__delattr__",
                        "__delete__",
                        "__delitem__",
                        "__dir__",
                        "__divmod__",
                        "__enter__",
                        "__eq__",
                        "__exit__",
                        "__float__",
                        "__floordiv__",
                        "__format__",
                        "__ge__",
                        "__get__",
                        "__getattr__",
                        "__getattribute__",
                        "__getitem__",
                        "__gt__",
                        "__hash__",
                        "__iadd__",
                        "__iand__",
                        "__ifloordiv__",
                        "__ilshift__",
                        "__imatmul__",
                        "__imod__",
                        "__imul__",
                        "__index__",
                        "__init__",
                        "__instancecheck__",
                        "__int__",
                        "__invert__",
                        "__ior__",
                        "__ipow__",
                        "__irshift__",
                        "__isub__",
                        "__iter__",
                        "__itruediv__",
                        "__ixor__",
                        "__le__",
                        "__len__",
                        "__length_hint__",
                        "__lshift__",
                        "__lt__",
                        "__matmul__",
                        "__missing__",
                        "__mod__",
                        "__mul__",
                        "__ne__",
                        "__neg__",
                        "__new__",
                        "__next__",
                        "__or__",
                        "__pos__",
                        "__pow__",
                        "__prepare__",
                        "__radd__",
                        "__rand__",
                        "__rdivmod__",
                        "__repr__",
                        "__reversed__",
                        "__rfloordiv__",
                        "__rlshift__",
                        "__rmatmul__",
                        "__rmod__",
                        "__rmul__",
                        "__ror__",
                        "__round__",
                        "__rpow__",
                        "__rrshift__",
                        "__rshift__",
                        "__rsub__",
                        "__rtruediv__",
                        "__rxor__",
                        "__set__",
                        "__setattr__",
                        "__setitem__",
                        "__str__",
                        "__sub__",
                        "__subclasscheck__",
                        "__truediv__",
                        "__xor__",
                    ),
                    suffix=r"\b",
                ),
                Name.Function.Magic,
            ),
        ],
        "magicvars": [
            (
                words(
                    (
                        "__annotations__",
                        "__bases__",
                        "__class__",
                        "__closure__",
                        "__code__",
                        "__defaults__",
                        "__dict__",
                        "__doc__",
                        "__file__",
                        "__func__",
                        "__globals__",
                        "__kwdefaults__",
                        "__module__",
                        "__mro__",
                        "__name__",
                        "__objclass__",
                        "__qualname__",
                        "__self__",
                        "__slots__",
                        "__weakref__",
                    ),
                    suffix=r"\b",
                ),
                Name.Variable.Magic,
            ),
        ],
        "numbers": [
            (
                r"(\d(?:_?\d)*\.(?:\d(?:_?\d)*)?|(?:\d(?:_?\d)*)?\.\d(?:_?\d)*)"
                r"([eE][+-]?\d(?:_?\d)*)?",
                Number.Float,
            ),
            (r"\d(?:_?\d)*[eE][+-]?\d(?:_?\d)*j?", Number.Float),
            (r"0[oO](?:_?[0-7])+", Number.Oct),
            (r"0[bB](?:_?[01])+", Number.Bin),
            (r"0[xX](?:_?[a-fA-F0-9])+", Number.Hex),
            (r"\d(?:_?\d)*", Number.Integer),
        ],
        "name": [
            (r"@" + uni_name, Name.Decorator),
            (r"@", Operator),  # new matrix multiplication operator
            (uni_name, Name),
        ],
        "funcname": [
            include("magicfuncs"),
            (uni_name, Name.Function, "#pop"),
            default("#pop"),
        ],
        "classname": [
            (uni_name, Name.Class, "#pop"),
        ],
        "import": [
            (r"(\s+)(as)(\s+)", bygroups(Text, Keyword, Text)),
            (r"\.", Name.Namespace),
            (uni_name, Name.Namespace),
            (r"(\s*)(,)(\s*)", bygroups(Text, Operator, Text)),
            default("#pop"),  # all else: go back
        ],
        "fromimport": [
            (r"(\s+)(import)\b", bygroups(Text, Keyword.Namespace), "#pop"),
            (r"\.", Name.Namespace),
            # if None occurs here, it's "raise x from None", since None can
            # never be a module name
            (r"None\b", Name.Builtin.Pseudo, "#pop"),
            (uni_name, Name.Namespace),
            default("#pop"),
        ],
        "rfstringescape": [
            (r"\{\{", String.Escape),
            (r"\}\}", String.Escape),
        ],
        "fstringescape": [
            include("rfstringescape"),
            include("stringescape"),
        ],
        "bytesescape": [
            (r'\\([\\abfnrtv"\']|\n|x[a-fA-F0-9]{2}|[0-7]{1,3})', String.Escape)
        ],
        "stringescape": [
            (r"\\(N\{.*?\}|u[a-fA-F0-9]{4}|U[a-fA-F0-9]{8})", String.Escape),
            include("bytesescape"),
        ],
        "fstrings-single": fstring_rules(String.Single),
        "fstrings-double": fstring_rules(String.Double),
        "strings-single": innerstring_rules(String.Single),
        "strings-double": innerstring_rules(String.Double),
        "dqf": [
            (r'"', String.Double, "#pop"),
            (r'\\\\|\\"|\\\n', String.Escape),  # included here for raw strings
            include("fstrings-double"),
        ],
        "sqf": [
            (r"'", String.Single, "#pop"),
            (r"\\\\|\\'|\\\n", String.Escape),  # included here for raw strings
            include("fstrings-single"),
        ],
        "dqs": [
            (r'"', String.Double, "#pop"),
            (r'\\\\|\\"|\\\n', String.Escape),  # included here for raw strings
            include("strings-double"),
        ],
        "sqs": [
            (r"'", String.Single, "#pop"),
            (r"\\\\|\\'|\\\n", String.Escape),  # included here for raw strings
            include("strings-single"),
        ],
        "tdqf": [
            (r'"""', String.Double, "#pop"),
            include("fstrings-double"),
            (r"\n", String.Double),
        ],
        "tsqf": [
            (r"'''", String.Single, "#pop"),
            include("fstrings-single"),
            (r"\n", String.Single),
        ],
        "tdqs": [
            (r'"""', String.Double, "#pop"),
            include("strings-double"),
            (r"\n", String.Double),
        ],
        "tsqs": [
            (r"'''", String.Single, "#pop"),
            include("strings-single"),
            (r"\n", String.Single),
        ],
    }

    def analyse_text(text):
        return shebang_matches(text, r"pythonw?(3(\.\d)?)?") or "import " in text[:1000]


Python3Lexer = PythonLexer


class Python2Lexer(RegexLexer):
    """
    For Python 2.x source code.

    .. versionchanged:: 2.5
       This class has been renamed from ``PythonLexer``.  ``PythonLexer`` now
       refers to the Python 3 variant.  File name patterns like ``*.py`` have
       been moved to Python 3 as well.
    """

    name = "Python 2.x"
    url = "http://www.python.org"
    aliases = ["python2", "py2"]
    filenames = []  # now taken over by PythonLexer (3.x)
    mimetypes = ["text/x-python2", "application/x-python2"]

    def innerstring_rules(ttype):
        return [
            # the old style '%s' % (...) string formatting
            (
                r"%(\(\w+\))?[-#0 +]*([0-9]+|[*])?(\.([0-9]+|[*]))?"
                "[hlL]?[E-GXc-giorsux%]",
                String.Interpol,
            ),
            # backslashes, quotes and formatting signs must be parsed one at a time
            (r'[^\\\'"%\n]+', ttype),
            (r'[\'"\\]', ttype),
            # unhandled string formatting sign
            (r"%", ttype),
            # newlines are an error (use "nl" state)
        ]

    tokens = {
        "root": [
            (r"\n", Text),
            (
                r'^(\s*)([rRuUbB]{,2})("""(?:.|\n)*?""")',
                bygroups(Text, String.Affix, String.Doc),
            ),
            (
                r"^(\s*)([rRuUbB]{,2})('''(?:.|\n)*?''')",
                bygroups(Text, String.Affix, String.Doc),
            ),
            (r"[^\S\n]+", Text),
            (r"\A#!.+$", Comment.Hashbang),
            (r"#.*$", Comment.Single),
            (r"[]{}:(),;[]", Punctuation),
            (r"\\\n", Text),
            (r"\\", Text),
            (r"(in|is|and|or|not)\b", Operator.Word),
            (r"!=|==|<<|>>|[-~+/*%=<>&^|.]", Operator),
            include("keywords"),
            (r"(def)((?:\s|\\\s)+)", bygroups(Keyword, Text), "funcname"),
            (r"(class)((?:\s|\\\s)+)", bygroups(Keyword, Text), "classname"),
            (r"(from)((?:\s|\\\s)+)", bygroups(Keyword.Namespace, Text), "fromimport"),
            (r"(import)((?:\s|\\\s)+)", bygroups(Keyword.Namespace, Text), "import"),
            include("builtins"),
            include("magicfuncs"),
            include("magicvars"),
            include("backtick"),
            (
                '([rR]|[uUbB][rR]|[rR][uUbB])(""")',
                bygroups(String.Affix, String.Double),
                "tdqs",
            ),
            (
                "([rR]|[uUbB][rR]|[rR][uUbB])(''')",
                bygroups(String.Affix, String.Single),
                "tsqs",
            ),
            (
                '([rR]|[uUbB][rR]|[rR][uUbB])(")',
                bygroups(String.Affix, String.Double),
                "dqs",
            ),
            (
                "([rR]|[uUbB][rR]|[rR][uUbB])(')",
                bygroups(String.Affix, String.Single),
                "sqs",
            ),
            (
                '([uUbB]?)(""")',
                bygroups(String.Affix, String.Double),
                combined("stringescape", "tdqs"),
            ),
            (
                "([uUbB]?)(''')",
                bygroups(String.Affix, String.Single),
                combined("stringescape", "tsqs"),
            ),
            (
                '([uUbB]?)(")',
                bygroups(String.Affix, String.Double),
                combined("stringescape", "dqs"),
            ),
            (
                "([uUbB]?)(')",
                bygroups(String.Affix, String.Single),
                combined("stringescape", "sqs"),
            ),
            include("name"),
            include("numbers"),
        ],
        "keywords": [
            (
                words(
                    (
                        "assert",
                        "break",
                        "continue",
                        "del",
                        "elif",
                        "else",
                        "except",
                        "exec",
                        "finally",
                        "for",
                        "global",
                        "if",
                        "lambda",
                        "pass",
                        "print",
                        "raise",
                        "return",
                        "try",
                        "while",
                        "yield",
                        "yield from",
                        "as",
                        "with",
                    ),
                    suffix=r"\b",
                ),
                Keyword,
            ),
        ],
        "builtins": [
            (
                words(
                    (
                        "__import__",
                        "abs",
                        "all",
                        "any",
                        "apply",
                        "basestring",
                        "bin",
                        "bool",
                        "buffer",
                        "bytearray",
                        "bytes",
                        "callable",
                        "chr",
                        "classmethod",
                        "cmp",
                        "coerce",
                        "compile",
                        "complex",
                        "delattr",
                        "dict",
                        "dir",
                        "divmod",
                        "enumerate",
                        "eval",
                        "execfile",
                        "exit",
                        "file",
                        "filter",
                        "float",
                        "frozenset",
                        "getattr",
                        "globals",
                        "hasattr",
                        "hash",
                        "hex",
                        "id",
                        "input",
                        "int",
                        "intern",
                        "isinstance",
                        "issubclass",
                        "iter",
                        "len",
                        "list",
                        "locals",
                        "long",
                        "map",
                        "max",
                        "min",
                        "next",
                        "object",
                        "oct",
                        "open",
                        "ord",
                        "pow",
                        "property",
                        "range",
                        "raw_input",
                        "reduce",
                        "reload",
                        "repr",
                        "reversed",
                        "round",
                        "set",
                        "setattr",
                        "slice",
                        "sorted",
                        "staticmethod",
                        "str",
                        "sum",
                        "super",
                        "tuple",
                        "type",
                        "unichr",
                        "unicode",
                        "vars",
                        "xrange",
                        "zip",
                    ),
                    prefix=r"(?<!\.)",
                    suffix=r"\b",
                ),
                Name.Builtin,
            ),
            (
                r"(?<!\.)(self|None|Ellipsis|NotImplemented|False|True|cls" r")\b",
                Name.Builtin.Pseudo,
            ),
            (
                words(
                    (
                        "ArithmeticError",
                        "AssertionError",
                        "AttributeError",
                        "BaseException",
                        "DeprecationWarning",
                        "EOFError",
                        "EnvironmentError",
                        "Exception",
                        "FloatingPointError",
                        "FutureWarning",
                        "GeneratorExit",
                        "IOError",
                        "ImportError",
                        "ImportWarning",
                        "IndentationError",
                        "IndexError",
                        "KeyError",
                        "KeyboardInterrupt",
                        "LookupError",
                        "MemoryError",
                        "NameError",
                        "NotImplementedError",
                        "OSError",
                        "OverflowError",
                        "OverflowWarning",
                        "PendingDeprecationWarning",
                        "ReferenceError",
                        "RuntimeError",
                        "RuntimeWarning",
                        "StandardError",
                        "StopIteration",
                        "SyntaxError",
                        "SyntaxWarning",
                        "SystemError",
                        "SystemExit",
                        "TabError",
                        "TypeError",
                        "UnboundLocalError",
                        "UnicodeDecodeError",
                        "UnicodeEncodeError",
                        "UnicodeError",
                        "UnicodeTranslateError",
                        "UnicodeWarning",
                        "UserWarning",
                        "ValueError",
                        "VMSError",
                        "Warning",
                        "WindowsError",
                        "ZeroDivisionError",
                    ),
                    prefix=r"(?<!\.)",
                    suffix=r"\b",
                ),
                Name.Exception,
            ),
        ],
        "magicfuncs": [
            (
                words(
                    (
                        "__abs__",
                        "__add__",
                        "__and__",
                        "__call__",
                        "__cmp__",
                        "__coerce__",
                        "__complex__",
                        "__contains__",
                        "__del__",
                        "__delattr__",
                        "__delete__",
                        "__delitem__",
                        "__delslice__",
                        "__div__",
                        "__divmod__",
                        "__enter__",
                        "__eq__",
                        "__exit__",
                        "__float__",
                        "__floordiv__",
                        "__ge__",
                        "__get__",
                        "__getattr__",
                        "__getattribute__",
                        "__getitem__",
                        "__getslice__",
                        "__gt__",
                        "__hash__",
                        "__hex__",
                        "__iadd__",
                        "__iand__",
                        "__idiv__",
                        "__ifloordiv__",
                        "__ilshift__",
                        "__imod__",
                        "__imul__",
                        "__index__",
                        "__init__",
                        "__instancecheck__",
                        "__int__",
                        "__invert__",
                        "__iop__",
                        "__ior__",
                        "__ipow__",
                        "__irshift__",
                        "__isub__",
                        "__iter__",
                        "__itruediv__",
                        "__ixor__",
                        "__le__",
                        "__len__",
                        "__long__",
                        "__lshift__",
                        "__lt__",
                        "__missing__",
                        "__mod__",
                        "__mul__",
                        "__ne__",
                        "__neg__",
                        "__new__",
                        "__nonzero__",
                        "__oct__",
                        "__op__",
                        "__or__",
                        "__pos__",
                        "__pow__",
                        "__radd__",
                        "__rand__",
                        "__rcmp__",
                        "__rdiv__",
                        "__rdivmod__",
                        "__repr__",
                        "__reversed__",
                        "__rfloordiv__",
                        "__rlshift__",
                        "__rmod__",
                        "__rmul__",
                        "__rop__",
                        "__ror__",
                        "__rpow__",
                        "__rrshift__",
                        "__rshift__",
                        "__rsub__",
                        "__rtruediv__",
                        "__rxor__",
                        "__set__",
                        "__setattr__",
                        "__setitem__",
                        "__setslice__",
                        "__str__",
                        "__sub__",
                        "__subclasscheck__",
                        "__truediv__",
                        "__unicode__",
                        "__xor__",
                    ),
                    suffix=r"\b",
                ),
                Name.Function.Magic,
            ),
        ],
        "magicvars": [
            (
                words(
                    (
                        "__bases__",
                        "__class__",
                        "__closure__",
                        "__code__",
                        "__defaults__",
                        "__dict__",
                        "__doc__",
                        "__file__",
                        "__func__",
                        "__globals__",
                        "__metaclass__",
                        "__module__",
                        "__mro__",
                        "__name__",
                        "__self__",
                        "__slots__",
                        "__weakref__",
                    ),
                    suffix=r"\b",
                ),
                Name.Variable.Magic,
            ),
        ],
        "numbers": [
            (r"(\d+\.\d*|\d*\.\d+)([eE][+-]?[0-9]+)?j?", Number.Float),
            (r"\d+[eE][+-]?[0-9]+j?", Number.Float),
            (r"0[0-7]+j?", Number.Oct),
            (r"0[bB][01]+", Number.Bin),
            (r"0[xX][a-fA-F0-9]+", Number.Hex),
            (r"\d+L", Number.Integer.Long),
            (r"\d+j?", Number.Integer),
        ],
        "backtick": [
            ("`.*?`", String.Backtick),
        ],
        "name": [
            (r"@[\w.]+", Name.Decorator),
            (r"[a-zA-Z_]\w*", Name),
        ],
        "funcname": [
            include("magicfuncs"),
            (r"[a-zA-Z_]\w*", Name.Function, "#pop"),
            default("#pop"),
        ],
        "classname": [(r"[a-zA-Z_]\w*", Name.Class, "#pop")],
        "import": [
            (r"(?:[ \t]|\\\n)+", Text),
            (r"as\b", Keyword.Namespace),
            (r",", Operator),
            (r"[a-zA-Z_][\w.]*", Name.Namespace),
            default("#pop"),  # all else: go back
        ],
        "fromimport": [
            (r"(?:[ \t]|\\\n)+", Text),
            (r"import\b", Keyword.Namespace, "#pop"),
            # if None occurs here, it's "raise x from None", since None can
            # never be a module name
            (r"None\b", Name.Builtin.Pseudo, "#pop"),
            # sadly, in "raise x from y" y will be highlighted as namespace too
            (r"[a-zA-Z_.][\w.]*", Name.Namespace),
            # anything else here also means "raise x from y" and is therefore
            # not an error
            default("#pop"),
        ],
        "stringescape": [
            (
                r'\\([\\abfnrtv"\']|\n|N\{.*?\}|u[a-fA-F0-9]{4}|'
                r"U[a-fA-F0-9]{8}|x[a-fA-F0-9]{2}|[0-7]{1,3})",
                String.Escape,
            )
        ],
        "strings-single": innerstring_rules(String.Single),
        "strings-double": innerstring_rules(String.Double),
        "dqs": [
            (r'"', String.Double, "#pop"),
            (r'\\\\|\\"|\\\n', String.Escape),  # included here for raw strings
            include("strings-double"),
        ],
        "sqs": [
            (r"'", String.Single, "#pop"),
            (r"\\\\|\\'|\\\n", String.Escape),  # included here for raw strings
            include("strings-single"),
        ],
        "tdqs": [
            (r'"""', String.Double, "#pop"),
            include("strings-double"),
            (r"\n", String.Double),
        ],
        "tsqs": [
            (r"'''", String.Single, "#pop"),
            include("strings-single"),
            (r"\n", String.Single),
        ],
    }

    def analyse_text(text):
        return shebang_matches(text, r"pythonw?2(\.\d)?")


class PythonConsoleLexer(Lexer):
    """
    For Python console output or doctests, such as:

    .. sourcecode:: pycon

        >>> a = 'foo'
        >>> print a
        foo
        >>> 1 / 0
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
        ZeroDivisionError: integer division or modulo by zero

    Additional options:

    `python3`
        Use Python 3 lexer for code.  Default is ``True``.

        .. versionadded:: 1.0
        .. versionchanged:: 2.5
           Now defaults to ``True``.
    """

    name = "Python console session"
    aliases = ["pycon"]
    mimetypes = ["text/x-python-doctest"]

    def __init__(self, **options):
        self.python3 = get_bool_opt(options, "python3", True)
        Lexer.__init__(self, **options)

    def get_tokens_unprocessed(self, text):
        if self.python3:
            pylexer = PythonLexer(**self.options)
            tblexer = PythonTracebackLexer(**self.options)
        else:
            pylexer = Python2Lexer(**self.options)
            tblexer = Python2TracebackLexer(**self.options)

        curcode = ""
        insertions = []
        curtb = ""
        tbindex = 0
        tb = 0
        for match in line_re.finditer(text):
            line = match.group()
            if line.startswith(">>> ") or line.startswith("... "):
                tb = 0
                insertions.append((len(curcode), [(0, Generic.Prompt, line[:4])]))
                curcode += line[4:]
            elif line.rstrip() == "..." and not tb:
                # only a new >>> prompt can end an exception block
                # otherwise an ellipsis in place of the traceback frames
                # will be mishandled
                insertions.append((len(curcode), [(0, Generic.Prompt, "...")]))
                curcode += line[3:]
            else:
                if curcode:
                    yield from do_insertions(
                        insertions, pylexer.get_tokens_unprocessed(curcode)
                    )
                    curcode = ""
                    insertions = []
                if line.startswith("Traceback (most recent call last):") or re.match(
                    '  File "[^"]+", line \\d+\\n$', line
                ):
                    tb = 1
                    curtb = line
                    tbindex = match.start()
                elif line == "KeyboardInterrupt\n":
                    yield match.start(), Name.Class, line
                elif tb:
                    curtb += line
                    if not (line.startswith(" ") or line.strip() == "..."):
                        tb = 0
                        for i, t, v in tblexer.get_tokens_unprocessed(curtb):
                            yield tbindex + i, t, v
                        curtb = ""
                else:
                    yield match.start(), Generic.Output, line
        if curcode:
            yield from do_insertions(
                insertions, pylexer.get_tokens_unprocessed(curcode)
            )
        if curtb:
            for i, t, v in tblexer.get_tokens_unprocessed(curtb):
                yield tbindex + i, t, v


class PythonTracebackLexer(RegexLexer):
    """
    For Python 3.x tracebacks, with support for chained exceptions.

    .. versionadded:: 1.0

    .. versionchanged:: 2.5
       This is now the default ``PythonTracebackLexer``.  It is still available
       as the alias ``Python3TracebackLexer``.
    """

    name = "Python Traceback"
    aliases = ["pytb", "py3tb"]
    filenames = ["*.pytb", "*.py3tb"]
    mimetypes = ["text/x-python-traceback", "text/x-python3-traceback"]

    tokens = {
        "root": [
            (r"\n", Text),
            (r"^Traceback \(most recent call last\):\n", Generic.Traceback, "intb"),
            (
                r"^During handling of the above exception, another "
                r"exception occurred:\n\n",
                Generic.Traceback,
            ),
            (
                r"^The above exception was the direct cause of the "
                r"following exception:\n\n",
                Generic.Traceback,
            ),
            (r'^(?=  File "[^"]+", line \d+)', Generic.Traceback, "intb"),
            (r"^.*\n", Other),
        ],
        "intb": [
            (
                r'^(  File )("[^"]+")(, line )(\d+)(, in )(.+)(\n)',
                bygroups(Text, Name.Builtin, Text, Number, Text, Name, Text),
            ),
            (
                r'^(  File )("[^"]+")(, line )(\d+)(\n)',
                bygroups(Text, Name.Builtin, Text, Number, Text),
            ),
            (r"^(    )(.+)(\n)", bygroups(Text, using(PythonLexer), Text), "markers"),
            (
                r"^([ \t]*)(\.\.\.)(\n)",
                bygroups(Text, Comment, Text),
            ),  # for doctests...
            (
                r"^([^:]+)(: )(.+)(\n)",
                bygroups(Generic.Error, Text, Name, Text),
                "#pop",
            ),
            (r"^([a-zA-Z_][\w.]*)(:?\n)", bygroups(Generic.Error, Text), "#pop"),
        ],
        "markers": [
            # Either `PEP 657 <https://www.python.org/dev/peps/pep-0657/>`
            # error locations in Python 3.11+, or single-caret markers
            # for syntax errors before that.
            (r"^( {4,})([~^]+)(\n)", bygroups(Text, Punctuation.Marker, Text), "#pop"),
            default("#pop"),
        ],
    }


Python3TracebackLexer = PythonTracebackLexer


class Python2TracebackLexer(RegexLexer):
    """
    For Python tracebacks.

    .. versionadded:: 0.7

    .. versionchanged:: 2.5
       This class has been renamed from ``PythonTracebackLexer``.
       ``PythonTracebackLexer`` now refers to the Python 3 variant.
    """

    name = "Python 2.x Traceback"
    aliases = ["py2tb"]
    filenames = ["*.py2tb"]
    mimetypes = ["text/x-python2-traceback"]

    tokens = {
        "root": [
            # Cover both (most recent call last) and (innermost last)
            # The optional ^C allows us to catch keyboard interrupt signals.
            (r"^(\^C)?(Traceback.*\n)", bygroups(Text, Generic.Traceback), "intb"),
            # SyntaxError starts with this.
            (r'^(?=  File "[^"]+", line \d+)', Generic.Traceback, "intb"),
            (r"^.*\n", Other),
        ],
        "intb": [
            (
                r'^(  File )("[^"]+")(, line )(\d+)(, in )(.+)(\n)',
                bygroups(Text, Name.Builtin, Text, Number, Text, Name, Text),
            ),
            (
                r'^(  File )("[^"]+")(, line )(\d+)(\n)',
                bygroups(Text, Name.Builtin, Text, Number, Text),
            ),
            (r"^(    )(.+)(\n)", bygroups(Text, using(Python2Lexer), Text), "marker"),
            (
                r"^([ \t]*)(\.\.\.)(\n)",
                bygroups(Text, Comment, Text),
            ),  # for doctests...
            (
                r"^([^:]+)(: )(.+)(\n)",
                bygroups(Generic.Error, Text, Name, Text),
                "#pop",
            ),
            (r"^([a-zA-Z_]\w*)(:?\n)", bygroups(Generic.Error, Text), "#pop"),
        ],
        "marker": [
            # For syntax errors.
            (r"( {4,})(\^)", bygroups(Text, Punctuation.Marker), "#pop"),
            default("#pop"),
        ],
    }


class CythonLexer(RegexLexer):
    """
    For Pyrex and Cython source code.

    .. versionadded:: 1.1
    """

    name = "Cython"
    url = "http://cython.org"
    aliases = ["cython", "pyx", "pyrex"]
    filenames = ["*.pyx", "*.pxd", "*.pxi"]
    mimetypes = ["text/x-cython", "application/x-cython"]

    tokens = {
        "root": [
            (r"\n", Text),
            (r'^(\s*)("""(?:.|\n)*?""")', bygroups(Text, String.Doc)),
            (r"^(\s*)('''(?:.|\n)*?''')", bygroups(Text, String.Doc)),
            (r"[^\S\n]+", Text),
            (r"#.*$", Comment),
            (r"[]{}:(),;[]", Punctuation),
            (r"\\\n", Text),
            (r"\\", Text),
            (r"(in|is|and|or|not)\b", Operator.Word),
            (
                r"(<)([a-zA-Z0-9.?]+)(>)",
                bygroups(Punctuation, Keyword.Type, Punctuation),
            ),
            (r"!=|==|<<|>>|[-~+/*%=<>&^|.?]", Operator),
            (
                r"(from)(\d+)(<=)(\s+)(<)(\d+)(:)",
                bygroups(
                    Keyword, Number.Integer, Operator, Name, Operator, Name, Punctuation
                ),
            ),
            include("keywords"),
            (r"(def|property)(\s+)", bygroups(Keyword, Text), "funcname"),
            (r"(cp?def)(\s+)", bygroups(Keyword, Text), "cdef"),
            # (should actually start a block with only cdefs)
            (r"(cdef)(:)", bygroups(Keyword, Punctuation)),
            (r"(class|struct)(\s+)", bygroups(Keyword, Text), "classname"),
            (r"(from)(\s+)", bygroups(Keyword, Text), "fromimport"),
            (r"(c?import)(\s+)", bygroups(Keyword, Text), "import"),
            include("builtins"),
            include("backtick"),
            ('(?:[rR]|[uU][rR]|[rR][uU])"""', String, "tdqs"),
            ("(?:[rR]|[uU][rR]|[rR][uU])'''", String, "tsqs"),
            ('(?:[rR]|[uU][rR]|[rR][uU])"', String, "dqs"),
            ("(?:[rR]|[uU][rR]|[rR][uU])'", String, "sqs"),
            ('[uU]?"""', String, combined("stringescape", "tdqs")),
            ("[uU]?'''", String, combined("stringescape", "tsqs")),
            ('[uU]?"', String, combined("stringescape", "dqs")),
            ("[uU]?'", String, combined("stringescape", "sqs")),
            include("name"),
            include("numbers"),
        ],
        "keywords": [
            (
                words(
                    (
                        "assert",
                        "async",
                        "await",
                        "break",
                        "by",
                        "continue",
                        "ctypedef",
                        "del",
                        "elif",
                        "else",
                        "except",
                        "except?",
                        "exec",
                        "finally",
                        "for",
                        "fused",
                        "gil",
                        "global",
                        "if",
                        "include",
                        "lambda",
                        "nogil",
                        "pass",
                        "print",
                        "raise",
                        "return",
                        "try",
                        "while",
                        "yield",
                        "as",
                        "with",
                    ),
                    suffix=r"\b",
                ),
                Keyword,
            ),
            (r"(DEF|IF|ELIF|ELSE)\b", Comment.Preproc),
        ],
        "builtins": [
            (
                words(
                    (
                        "__import__",
                        "abs",
                        "all",
                        "any",
                        "apply",
                        "basestring",
                        "bin",
                        "bint",
                        "bool",
                        "buffer",
                        "bytearray",
                        "bytes",
                        "callable",
                        "chr",
                        "classmethod",
                        "cmp",
                        "coerce",
                        "compile",
                        "complex",
                        "delattr",
                        "dict",
                        "dir",
                        "divmod",
                        "enumerate",
                        "eval",
                        "execfile",
                        "exit",
                        "file",
                        "filter",
                        "float",
                        "frozenset",
                        "getattr",
                        "globals",
                        "hasattr",
                        "hash",
                        "hex",
                        "id",
                        "input",
                        "int",
                        "intern",
                        "isinstance",
                        "issubclass",
                        "iter",
                        "len",
                        "list",
                        "locals",
                        "long",
                        "map",
                        "max",
                        "min",
                        "next",
                        "object",
                        "oct",
                        "open",
                        "ord",
                        "pow",
                        "property",
                        "Py_ssize_t",
                        "range",
                        "raw_input",
                        "reduce",
                        "reload",
                        "repr",
                        "reversed",
                        "round",
                        "set",
                        "setattr",
                        "slice",
                        "sorted",
                        "staticmethod",
                        "str",
                        "sum",
                        "super",
                        "tuple",
                        "type",
                        "unichr",
                        "unicode",
                        "unsigned",
                        "vars",
                        "xrange",
                        "zip",
                    ),
                    prefix=r"(?<!\.)",
                    suffix=r"\b",
                ),
                Name.Builtin,
            ),
            (
                r"(?<!\.)(self|None|Ellipsis|NotImplemented|False|True|NULL" r")\b",
                Name.Builtin.Pseudo,
            ),
            (
                words(
                    (
                        "ArithmeticError",
                        "AssertionError",
                        "AttributeError",
                        "BaseException",
                        "DeprecationWarning",
                        "EOFError",
                        "EnvironmentError",
                        "Exception",
                        "FloatingPointError",
                        "FutureWarning",
                        "GeneratorExit",
                        "IOError",
                        "ImportError",
                        "ImportWarning",
                        "IndentationError",
                        "IndexError",
                        "KeyError",
                        "KeyboardInterrupt",
                        "LookupError",
                        "MemoryError",
                        "NameError",
                        "NotImplemented",
                        "NotImplementedError",
                        "OSError",
                        "OverflowError",
                        "OverflowWarning",
                        "PendingDeprecationWarning",
                        "ReferenceError",
                        "RuntimeError",
                        "RuntimeWarning",
                        "StandardError",
                        "StopIteration",
                        "SyntaxError",
                        "SyntaxWarning",
                        "SystemError",
                        "SystemExit",
                        "TabError",
                        "TypeError",
                        "UnboundLocalError",
                        "UnicodeDecodeError",
                        "UnicodeEncodeError",
                        "UnicodeError",
                        "UnicodeTranslateError",
                        "UnicodeWarning",
                        "UserWarning",
                        "ValueError",
                        "Warning",
                        "ZeroDivisionError",
                    ),
                    prefix=r"(?<!\.)",
                    suffix=r"\b",
                ),
                Name.Exception,
            ),
        ],
        "numbers": [
            (r"(\d+\.?\d*|\d*\.\d+)([eE][+-]?[0-9]+)?", Number.Float),
            (r"0\d+", Number.Oct),
            (r"0[xX][a-fA-F0-9]+", Number.Hex),
            (r"\d+L", Number.Integer.Long),
            (r"\d+", Number.Integer),
        ],
        "backtick": [
            ("`.*?`", String.Backtick),
        ],
        "name": [
            (r"@\w+", Name.Decorator),
            (r"[a-zA-Z_]\w*", Name),
        ],
        "funcname": [(r"[a-zA-Z_]\w*", Name.Function, "#pop")],
        "cdef": [
            (r"(public|readonly|extern|api|inline)\b", Keyword.Reserved),
            (r"(struct|enum|union|class)\b", Keyword),
            (r"([a-zA-Z_]\w*)(\s*)(?=[(:#=]|$)", bygroups(Name.Function, Text), "#pop"),
            (r"([a-zA-Z_]\w*)(\s*)(,)", bygroups(Name.Function, Text, Punctuation)),
            (r"from\b", Keyword, "#pop"),
            (r"as\b", Keyword),
            (r":", Punctuation, "#pop"),
            (r'(?=["\'])', Text, "#pop"),
            (r"[a-zA-Z_]\w*", Keyword.Type),
            (r".", Text),
        ],
        "classname": [(r"[a-zA-Z_]\w*", Name.Class, "#pop")],
        "import": [
            (r"(\s+)(as)(\s+)", bygroups(Text, Keyword, Text)),
            (r"[a-zA-Z_][\w.]*", Name.Namespace),
            (r"(\s*)(,)(\s*)", bygroups(Text, Operator, Text)),
            default("#pop"),  # all else: go back
        ],
        "fromimport": [
            (r"(\s+)(c?import)\b", bygroups(Text, Keyword), "#pop"),
            (r"[a-zA-Z_.][\w.]*", Name.Namespace),
            # ``cdef foo from "header"``, or ``for foo from 0 < i < 10``
            default("#pop"),
        ],
        "stringescape": [
            (
                r'\\([\\abfnrtv"\']|\n|N\{.*?\}|u[a-fA-F0-9]{4}|'
                r"U[a-fA-F0-9]{8}|x[a-fA-F0-9]{2}|[0-7]{1,3})",
                String.Escape,
            )
        ],
        "strings": [
            (
                r"%(\([a-zA-Z0-9]+\))?[-#0 +]*([0-9]+|[*])?(\.([0-9]+|[*]))?"
                "[hlL]?[E-GXc-giorsux%]",
                String.Interpol,
            ),
            (r'[^\\\'"%\n]+', String),
            # quotes, percents and backslashes must be parsed one at a time
            (r'[\'"\\]', String),
            # unhandled string formatting sign
            (r"%", String)
            # newlines are an error (use "nl" state)
        ],
        "nl": [(r"\n", String)],
        "dqs": [
            (r'"', String, "#pop"),
            (r'\\\\|\\"|\\\n', String.Escape),  # included here again for raw strings
            include("strings"),
        ],
        "sqs": [
            (r"'", String, "#pop"),
            (r"\\\\|\\'|\\\n", String.Escape),  # included here again for raw strings
            include("strings"),
        ],
        "tdqs": [(r'"""', String, "#pop"), include("strings"), include("nl")],
        "tsqs": [(r"'''", String, "#pop"), include("strings"), include("nl")],
    }


class DgLexer(RegexLexer):
    """
    Lexer for dg,
    a functional and object-oriented programming language
    running on the CPython 3 VM.

    .. versionadded:: 1.6
    """

    name = "dg"
    aliases = ["dg"]
    filenames = ["*.dg"]
    mimetypes = ["text/x-dg"]

    tokens = {
        "root": [
            (r"\s+", Text),
            (r"#.*?$", Comment.Single),
            (r"(?i)0b[01]+", Number.Bin),
            (r"(?i)0o[0-7]+", Number.Oct),
            (r"(?i)0x[0-9a-f]+", Number.Hex),
            (r"(?i)[+-]?[0-9]+\.[0-9]+(e[+-]?[0-9]+)?j?", Number.Float),
            (r"(?i)[+-]?[0-9]+e[+-]?\d+j?", Number.Float),
            (r"(?i)[+-]?[0-9]+j?", Number.Integer),
            (r"(?i)(br|r?b?)'''", String, combined("stringescape", "tsqs", "string")),
            (r'(?i)(br|r?b?)"""', String, combined("stringescape", "tdqs", "string")),
            (r"(?i)(br|r?b?)'", String, combined("stringescape", "sqs", "string")),
            (r'(?i)(br|r?b?)"', String, combined("stringescape", "dqs", "string")),
            (r"`\w+'*`", Operator),
            (r"\b(and|in|is|or|where)\b", Operator.Word),
            (r"[!$%&*+\-./:<-@\\^|~;,]+", Operator),
            (
                words(
                    (
                        "bool",
                        "bytearray",
                        "bytes",
                        "classmethod",
                        "complex",
                        "dict",
                        "dict'",
                        "float",
                        "frozenset",
                        "int",
                        "list",
                        "list'",
                        "memoryview",
                        "object",
                        "property",
                        "range",
                        "set",
                        "set'",
                        "slice",
                        "staticmethod",
                        "str",
                        "super",
                        "tuple",
                        "tuple'",
                        "type",
                    ),
                    prefix=r"(?<!\.)",
                    suffix=r"(?![\'\w])",
                ),
                Name.Builtin,
            ),
            (
                words(
                    (
                        "__import__",
                        "abs",
                        "all",
                        "any",
                        "bin",
                        "bind",
                        "chr",
                        "cmp",
                        "compile",
                        "complex",
                        "delattr",
                        "dir",
                        "divmod",
                        "drop",
                        "dropwhile",
                        "enumerate",
                        "eval",
                        "exhaust",
                        "filter",
                        "flip",
                        "foldl1?",
                        "format",
                        "fst",
                        "getattr",
                        "globals",
                        "hasattr",
                        "hash",
                        "head",
                        "hex",
                        "id",
                        "init",
                        "input",
                        "isinstance",
                        "issubclass",
                        "iter",
                        "iterate",
                        "last",
                        "len",
                        "locals",
                        "map",
                        "max",
                        "min",
                        "next",
                        "oct",
                        "open",
                        "ord",
                        "pow",
                        "print",
                        "repr",
                        "reversed",
                        "round",
                        "setattr",
                        "scanl1?",
                        "snd",
                        "sorted",
                        "sum",
                        "tail",
                        "take",
                        "takewhile",
                        "vars",
                        "zip",
                    ),
                    prefix=r"(?<!\.)",
                    suffix=r"(?![\'\w])",
                ),
                Name.Builtin,
            ),
            (
                r"(?<!\.)(self|Ellipsis|NotImplemented|None|True|False)(?!['\w])",
                Name.Builtin.Pseudo,
            ),
            (r"(?<!\.)[A-Z]\w*(Error|Exception|Warning)'*(?!['\w])", Name.Exception),
            (
                r"(?<!\.)(Exception|GeneratorExit|KeyboardInterrupt|StopIteration|"
                r"SystemExit)(?!['\w])",
                Name.Exception,
            ),
            (
                r"(?<![\w.])(except|finally|for|if|import|not|otherwise|raise|"
                r"subclass|while|with|yield)(?!['\w])",
                Keyword.Reserved,
            ),
            (r"[A-Z_]+'*(?!['\w])", Name),
            (r"[A-Z]\w+'*(?!['\w])", Keyword.Type),
            (r"\w+'*", Name),
            (r"[()]", Punctuation),
            (r".", Error),
        ],
        "stringescape": [
            (
                r'\\([\\abfnrtv"\']|\n|N\{.*?\}|u[a-fA-F0-9]{4}|'
                r"U[a-fA-F0-9]{8}|x[a-fA-F0-9]{2}|[0-7]{1,3})",
                String.Escape,
            )
        ],
        "string": [
            (
                r"%(\(\w+\))?[-#0 +]*([0-9]+|[*])?(\.([0-9]+|[*]))?"
                "[hlL]?[E-GXc-giorsux%]",
                String.Interpol,
            ),
            (r'[^\\\'"%\n]+', String),
            # quotes, percents and backslashes must be parsed one at a time
            (r'[\'"\\]', String),
            # unhandled string formatting sign
            (r"%", String),
            (r"\n", String),
        ],
        "dqs": [(r'"', String, "#pop")],
        "sqs": [(r"'", String, "#pop")],
        "tdqs": [(r'"""', String, "#pop")],
        "tsqs": [(r"'''", String, "#pop")],
    }


class NumPyLexer(PythonLexer):
    """
    A Python lexer recognizing Numerical Python builtins.

    .. versionadded:: 0.10
    """

    name = "NumPy"
    url = "https://numpy.org/"
    aliases = ["numpy"]

    # override the mimetypes to not inherit them from python
    mimetypes = []
    filenames = []

    EXTRA_KEYWORDS = {
        "abs",
        "absolute",
        "accumulate",
        "add",
        "alen",
        "all",
        "allclose",
        "alltrue",
        "alterdot",
        "amax",
        "amin",
        "angle",
        "any",
        "append",
        "apply_along_axis",
        "apply_over_axes",
        "arange",
        "arccos",
        "arccosh",
        "arcsin",
        "arcsinh",
        "arctan",
        "arctan2",
        "arctanh",
        "argmax",
        "argmin",
        "argsort",
        "argwhere",
        "around",
        "array",
        "array2string",
        "array_equal",
        "array_equiv",
        "array_repr",
        "array_split",
        "array_str",
        "arrayrange",
        "asanyarray",
        "asarray",
        "asarray_chkfinite",
        "ascontiguousarray",
        "asfarray",
        "asfortranarray",
        "asmatrix",
        "asscalar",
        "astype",
        "atleast_1d",
        "atleast_2d",
        "atleast_3d",
        "average",
        "bartlett",
        "base_repr",
        "beta",
        "binary_repr",
        "bincount",
        "binomial",
        "bitwise_and",
        "bitwise_not",
        "bitwise_or",
        "bitwise_xor",
        "blackman",
        "bmat",
        "broadcast",
        "byte_bounds",
        "bytes",
        "byteswap",
        "c_",
        "can_cast",
        "ceil",
        "choose",
        "clip",
        "column_stack",
        "common_type",
        "compare_chararrays",
        "compress",
        "concatenate",
        "conj",
        "conjugate",
        "convolve",
        "copy",
        "corrcoef",
        "correlate",
        "cos",
        "cosh",
        "cov",
        "cross",
        "cumprod",
        "cumproduct",
        "cumsum",
        "delete",
        "deprecate",
        "diag",
        "diagflat",
        "diagonal",
        "diff",
        "digitize",
        "disp",
        "divide",
        "dot",
        "dsplit",
        "dstack",
        "dtype",
        "dump",
        "dumps",
        "ediff1d",
        "empty",
        "empty_like",
        "equal",
        "exp",
        "expand_dims",
        "expm1",
        "extract",
        "eye",
        "fabs",
        "fastCopyAndTranspose",
        "fft",
        "fftfreq",
        "fftshift",
        "fill",
        "finfo",
        "fix",
        "flat",
        "flatnonzero",
        "flatten",
        "fliplr",
        "flipud",
        "floor",
        "floor_divide",
        "fmod",
        "frexp",
        "fromarrays",
        "frombuffer",
        "fromfile",
        "fromfunction",
        "fromiter",
        "frompyfunc",
        "fromstring",
        "generic",
        "get_array_wrap",
        "get_include",
        "get_numarray_include",
        "get_numpy_include",
        "get_printoptions",
        "getbuffer",
        "getbufsize",
        "geterr",
        "geterrcall",
        "geterrobj",
        "getfield",
        "gradient",
        "greater",
        "greater_equal",
        "gumbel",
        "hamming",
        "hanning",
        "histogram",
        "histogram2d",
        "histogramdd",
        "hsplit",
        "hstack",
        "hypot",
        "i0",
        "identity",
        "ifft",
        "imag",
        "index_exp",
        "indices",
        "inf",
        "info",
        "inner",
        "insert",
        "int_asbuffer",
        "interp",
        "intersect1d",
        "intersect1d_nu",
        "inv",
        "invert",
        "iscomplex",
        "iscomplexobj",
        "isfinite",
        "isfortran",
        "isinf",
        "isnan",
        "isneginf",
        "isposinf",
        "isreal",
        "isrealobj",
        "isscalar",
        "issctype",
        "issubclass_",
        "issubdtype",
        "issubsctype",
        "item",
        "itemset",
        "iterable",
        "ix_",
        "kaiser",
        "kron",
        "ldexp",
        "left_shift",
        "less",
        "less_equal",
        "lexsort",
        "linspace",
        "load",
        "loads",
        "loadtxt",
        "log",
        "log10",
        "log1p",
        "log2",
        "logical_and",
        "logical_not",
        "logical_or",
        "logical_xor",
        "logspace",
        "lstsq",
        "mat",
        "matrix",
        "max",
        "maximum",
        "maximum_sctype",
        "may_share_memory",
        "mean",
        "median",
        "meshgrid",
        "mgrid",
        "min",
        "minimum",
        "mintypecode",
        "mod",
        "modf",
        "msort",
        "multiply",
        "nan",
        "nan_to_num",
        "nanargmax",
        "nanargmin",
        "nanmax",
        "nanmin",
        "nansum",
        "ndenumerate",
        "ndim",
        "ndindex",
        "negative",
        "newaxis",
        "newbuffer",
        "newbyteorder",
        "nonzero",
        "not_equal",
        "obj2sctype",
        "ogrid",
        "ones",
        "ones_like",
        "outer",
        "permutation",
        "piecewise",
        "pinv",
        "pkgload",
        "place",
        "poisson",
        "poly",
        "poly1d",
        "polyadd",
        "polyder",
        "polydiv",
        "polyfit",
        "polyint",
        "polymul",
        "polysub",
        "polyval",
        "power",
        "prod",
        "product",
        "ptp",
        "put",
        "putmask",
        "r_",
        "randint",
        "random_integers",
        "random_sample",
        "ranf",
        "rank",
        "ravel",
        "real",
        "real_if_close",
        "recarray",
        "reciprocal",
        "reduce",
        "remainder",
        "repeat",
        "require",
        "reshape",
        "resize",
        "restoredot",
        "right_shift",
        "rint",
        "roll",
        "rollaxis",
        "roots",
        "rot90",
        "round",
        "round_",
        "row_stack",
        "s_",
        "sample",
        "savetxt",
        "sctype2char",
        "searchsorted",
        "seed",
        "select",
        "set_numeric_ops",
        "set_printoptions",
        "set_string_function",
        "setbufsize",
        "setdiff1d",
        "seterr",
        "seterrcall",
        "seterrobj",
        "setfield",
        "setflags",
        "setmember1d",
        "setxor1d",
        "shape",
        "show_config",
        "shuffle",
        "sign",
        "signbit",
        "sin",
        "sinc",
        "sinh",
        "size",
        "slice",
        "solve",
        "sometrue",
        "sort",
        "sort_complex",
        "source",
        "split",
        "sqrt",
        "square",
        "squeeze",
        "standard_normal",
        "std",
        "subtract",
        "sum",
        "svd",
        "swapaxes",
        "take",
        "tan",
        "tanh",
        "tensordot",
        "test",
        "tile",
        "tofile",
        "tolist",
        "tostring",
        "trace",
        "transpose",
        "trapz",
        "tri",
        "tril",
        "trim_zeros",
        "triu",
        "true_divide",
        "typeDict",
        "typename",
        "uniform",
        "union1d",
        "unique",
        "unique1d",
        "unravel_index",
        "unwrap",
        "vander",
        "var",
        "vdot",
        "vectorize",
        "view",
        "vonmises",
        "vsplit",
        "vstack",
        "weibull",
        "where",
        "who",
        "zeros",
        "zeros_like",
    }

    def get_tokens_unprocessed(self, text):
        for index, token, value in PythonLexer.get_tokens_unprocessed(self, text):
            if token is Name and value in self.EXTRA_KEYWORDS:
                yield index, Keyword.Pseudo, value
            else:
                yield index, token, value

    def analyse_text(text):
        ltext = text[:1000]
        return (
            shebang_matches(text, r"pythonw?(3(\.\d)?)?") or "import " in ltext
        ) and ("import numpy" in ltext or "from numpy import" in ltext)



"""
This module implements Git's wildmatch pattern matching which itself is
derived from Rsync's wildmatch. Git uses wildmatch for its ".gitignore"
files.
"""

import re
import warnings
from typing import AnyStr, Optional, Tuple

from .. import util
from ..pattern import RegexPattern

_BYTES_ENCODING = "latin1"
"""
The encoding to use when parsing a byte string pattern.
"""

_DIR_MARK = "ps_d"
"""
The regex group name for the directory marker. This is only used by
:class:`GitIgnoreSpec`.
"""


class GitWildMatchPatternError(ValueError):
    """
    The :class:`GitWildMatchPatternError` indicates an invalid git wild match
    pattern.
    """

    pass


class GitWildMatchPattern(RegexPattern):
    """
    The :class:`GitWildMatchPattern` class represents a compiled Git
    wildmatch pattern.
    """

    # Keep the dict-less class hierarchy.
    __slots__ = ()

    @classmethod
    def pattern_to_regex(
        cls,
        pattern: AnyStr,
    ) -> Tuple[Optional[AnyStr], Optional[bool]]:
        """
        Convert the pattern into a regular expression.

        *pattern* (:class:`str` or :class:`bytes`) is the pattern to convert
        into a regular expression.

        Returns the uncompiled regular expression (:class:`str`, :class:`bytes`,
        or :data:`None`); and whether matched files should be included
        (:data:`True`), excluded (:data:`False`), or if it is a
        null-operation (:data:`None`).
        """
        if isinstance(pattern, str):
            return_type = str
        elif isinstance(pattern, bytes):
            return_type = bytes
            pattern = pattern.decode(_BYTES_ENCODING)
        else:
            raise TypeError(f"pattern:{pattern!r} is not a unicode or byte string.")

        original_pattern = pattern
        pattern = pattern.strip()

        if pattern.startswith("#"):
            # A pattern starting with a hash ('#') serves as a comment
            # (neither includes nor excludes files). Escape the hash with a
            # back-slash to match a literal hash (i.e., '\#').
            regex = None
            include = None

        elif pattern == "/":
            # EDGE CASE: According to `git check-ignore` (v2.4.1), a single
            # '/' does not match any file.
            regex = None
            include = None

        elif pattern:
            if pattern.startswith("!"):
                # A pattern starting with an exclamation mark ('!') negates the
                # pattern (exclude instead of include). Escape the exclamation
                # mark with a back-slash to match a literal exclamation mark
                # (i.e., '\!').
                include = False
                # Remove leading exclamation mark.
                pattern = pattern[1:]
            else:
                include = True

            # Allow a regex override for edge cases that cannot be handled
            # through normalization.
            override_regex = None

            # Split pattern into segments.
            pattern_segs = pattern.split("/")

            # Normalize pattern to make processing easier.

            # EDGE CASE: Deal with duplicate double-asterisk sequences.
            # Collapse each sequence down to one double-asterisk. Iterate over
            # the segments in reverse and remove the duplicate double
            # asterisks as we go.
            for i in range(len(pattern_segs) - 1, 0, -1):
                prev = pattern_segs[i - 1]
                seg = pattern_segs[i]
                if prev == "**" and seg == "**":
                    del pattern_segs[i]

            if (
                len(pattern_segs) == 2
                and pattern_segs[0] == "**"
                and not pattern_segs[1]
            ):
                # EDGE CASE: The '**/' pattern should match everything except
                # individual files in the root directory. This case cannot be
                # adequately handled through normalization. Use the override.
                override_regex = f"^.+(?P<{_DIR_MARK}>/).*$"

            if not pattern_segs[0]:
                # A pattern beginning with a slash ('/') will only match paths
                # directly on the root directory instead of any descendant
                # paths. So, remove empty first segment to make pattern relative
                # to root.
                del pattern_segs[0]

            elif len(pattern_segs) == 1 or (
                len(pattern_segs) == 2 and not pattern_segs[1]
            ):
                # A single pattern without a beginning slash ('/') will match
                # any descendant path. This is equivalent to "**/{pattern}". So,
                # prepend with double-asterisks to make pattern relative to
                # root.
                # EDGE CASE: This also holds for a single pattern with a
                # trailing slash (e.g. dir/).
                if pattern_segs[0] != "**":
                    pattern_segs.insert(0, "**")

            else:
                # EDGE CASE: A pattern without a beginning slash ('/') but
                # contains at least one prepended directory (e.g.
                # "dir/{pattern}") should not match "**/dir/{pattern}",
                # according to `git check-ignore` (v2.4.1).
                pass

            if not pattern_segs:
                # After resolving the edge cases, we end up with no pattern at
                # all. This must be because the pattern is invalid.
                raise GitWildMatchPatternError(
                    f"Invalid git pattern: {original_pattern!r}"
                )

            if not pattern_segs[-1] and len(pattern_segs) > 1:
                # A pattern ending with a slash ('/') will match all descendant
                # paths if it is a directory but not if it is a regular file.
                # This is equivalent to "{pattern}/**". So, set last segment to
                # a double-asterisk to include all descendants.
                pattern_segs[-1] = "**"

            if override_regex is None:
                # Build regular expression from pattern.
                output = ["^"]
                need_slash = False
                end = len(pattern_segs) - 1
                for i, seg in enumerate(pattern_segs):
                    if seg == "**":
                        if i == 0 and i == end:
                            # A pattern consisting solely of double-asterisks ('**')
                            # will match every path.
                            output.append(f"[^/]+(?:(?P<{_DIR_MARK}>/).*)?")
                        elif i == 0:
                            # A normalized pattern beginning with double-asterisks
                            # ('**') will match any leading path segments.
                            output.append("(?:.+/)?")
                            need_slash = False
                        elif i == end:
                            # A normalized pattern ending with double-asterisks ('**')
                            # will match any trailing path segments.
                            output.append(f"(?P<{_DIR_MARK}>/).*")
                        else:
                            # A pattern with inner double-asterisks ('**') will match
                            # multiple (or zero) inner path segments.
                            output.append("(?:/.+)?")
                            need_slash = True

                    elif seg == "*":
                        # Match single path segment.
                        if need_slash:
                            output.append("/")

                        output.append("[^/]+")

                        if i == end:
                            # A pattern ending without a slash ('/') will match a file
                            # or a directory (with paths underneath it). E.g., "foo"
                            # matches "foo", "foo/bar", "foo/bar/baz", etc.
                            output.append(f"(?:(?P<{_DIR_MARK}>/).*)?")

                        need_slash = True

                    else:
                        # Match segment glob pattern.
                        if need_slash:
                            output.append("/")

                        try:
                            output.append(cls._translate_segment_glob(seg))
                        except ValueError as e:
                            raise GitWildMatchPatternError(
                                f"Invalid git pattern: {original_pattern!r}"
                            ) from e

                        if i == end:
                            # A pattern ending without a slash ('/') will match a file
                            # or a directory (with paths underneath it). E.g., "foo"
                            # matches "foo", "foo/bar", "foo/bar/baz", etc.
                            output.append(f"(?:(?P<{_DIR_MARK}>/).*)?")

                        need_slash = True

                output.append("$")
                regex = "".join(output)

            else:
                # Use regex override.
                regex = override_regex

        else:
            # A blank pattern is a null-operation (neither includes nor
            # excludes files).
            regex = None
            include = None

        if regex is not None and return_type is bytes:
            regex = regex.encode(_BYTES_ENCODING)

        return regex, include

    @staticmethod
    def _translate_segment_glob(pattern: str) -> str:
        """
        Translates the glob pattern to a regular expression. This is used in
        the constructor to translate a path segment glob pattern to its
        corresponding regular expression.

        *pattern* (:class:`str`) is the glob pattern.

        Returns the regular expression (:class:`str`).
        """
        # NOTE: This is derived from `fnmatch.translate()` and is similar to
        # the POSIX function `fnmatch()` with the `FNM_PATHNAME` flag set.

        escape = False
        regex = ""
        i, end = 0, len(pattern)
        while i < end:
            # Get next character.
            char = pattern[i]
            i += 1

            if escape:
                # Escape the character.
                escape = False
                regex += re.escape(char)

            elif char == "\\":
                # Escape character, escape next character.
                escape = True

            elif char == "*":
                # Multi-character wildcard. Match any string (except slashes),
                # including an empty string.
                regex += "[^/]*"

            elif char == "?":
                # Single-character wildcard. Match any single character (except
                # a slash).
                regex += "[^/]"

            elif char == "[":
                # Bracket expression wildcard. Except for the beginning
                # exclamation mark, the whole bracket expression can be used
                # directly as regex but we have to find where the expression
                # ends.
                # - "[][!]" matches ']', '[' and '!'.
                # - "[]-]" matches ']' and '-'.
                # - "[!]a-]" matches any character except ']', 'a' and '-'.
                j = i
                # Pass brack expression negation.
                if j < end and pattern[j] == "!":
                    j += 1
                # Pass first closing bracket if it is at the beginning of the
                # expression.
                if j < end and pattern[j] == "]":
                    j += 1
                # Find closing bracket. Stop once we reach the end or find it.
                while j < end and pattern[j] != "]":
                    j += 1

                if j < end:
                    # Found end of bracket expression. Increment j to be one past
                    # the closing bracket:
                    #
                    #  [...]
                    #   ^   ^
                    #   i   j
                    #
                    j += 1
                    expr = "["

                    if pattern[i] == "!":
                        # Braket expression needs to be negated.
                        expr += "^"
                        i += 1
                    elif pattern[i] == "^":
                        # POSIX declares that the regex bracket expression negation
                        # "[^...]" is undefined in a glob pattern. Python's
                        # `fnmatch.translate()` escapes the caret ('^') as a
                        # literal. To maintain consistency with undefined behavior,
                        # I am escaping the '^' as well.
                        expr += "\\^"
                        i += 1

                    # Build regex bracket expression. Escape slashes so they are
                    # treated as literal slashes by regex as defined by POSIX.
                    expr += pattern[i:j].replace("\\", "\\\\")

                    # Add regex bracket expression to regex result.
                    regex += expr

                    # Set i to one past the closing bracket.
                    i = j

                else:
                    # Failed to find closing bracket, treat opening bracket as a
                    # bracket literal instead of as an expression.
                    regex += "\\["

            else:
                # Regular character, escape it for regex.
                regex += re.escape(char)

        if escape:
            raise ValueError(
                f"Escape character found with no next character to escape: {pattern!r}"
            )

        return regex

    @staticmethod
    def escape(s: AnyStr) -> AnyStr:
        """
        Escape special characters in the given string.

        *s* (:class:`str` or :class:`bytes`) a filename or a string that you
        want to escape, usually before adding it to a ".gitignore".

        Returns the escaped string (:class:`str` or :class:`bytes`).
        """
        if isinstance(s, str):
            return_type = str
            string = s
        elif isinstance(s, bytes):
            return_type = bytes
            string = s.decode(_BYTES_ENCODING)
        else:
            raise TypeError(f"s:{s!r} is not a unicode or byte string.")

        # Reference: https://git-scm.com/docs/gitignore#_pattern_format
        meta_characters = r"[]!*#?"

        out_string = "".join("\\" + x if x in meta_characters else x for x in string)

        if return_type is bytes:
            return out_string.encode(_BYTES_ENCODING)
        else:
            return out_string


util.register_pattern("gitwildmatch", GitWildMatchPattern)


class GitIgnorePattern(GitWildMatchPattern):
    """
    The :class:`GitIgnorePattern` class is deprecated by :class:`GitWildMatchPattern`.
    This class only exists to maintain compatibility with v0.4.
    """

    def __init__(self, *args, **kw) -> None:
        """
        Warn about deprecation.
        """
        self._deprecated()
        super(GitIgnorePattern, self).__init__(*args, **kw)

    @staticmethod
    def _deprecated() -> None:
        """
        Warn about deprecation.
        """
        warnings.warn(
            (
                "GitIgnorePattern ('gitignore') is deprecated. Use "
                "GitWildMatchPattern ('gitwildmatch') instead."
            ),
            DeprecationWarning,
            stacklevel=3,
        )

    @classmethod
    def pattern_to_regex(cls, *args, **kw):
        """
        Warn about deprecation.
        """
        cls._deprecated()
        return super(GitIgnorePattern, cls).pattern_to_regex(*args, **kw)


# Register `GitIgnorePattern` as "gitignore" for backward compatibility
# with v0.4.
util.register_pattern("gitignore", GitIgnorePattern)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2020 Vinay Sajip.
# Licensed to the Python Software Foundation under a contributor agreement.
# See LICENSE.txt and CONTRIBUTORS.txt.
#
from __future__ import unicode_literals

import base64
import codecs
import datetime
from email import message_from_file
import hashlib
import json
import logging
import os
import posixpath
import re
import shutil
import sys
import tempfile
import zipfile

from . import __version__, DistlibException
from .compat import sysconfig, ZipFile, fsdecode, text_type, filter
from .database import InstalledDistribution
from .metadata import (
    Metadata,
    METADATA_FILENAME,
    WHEEL_METADATA_FILENAME,
    LEGACY_METADATA_FILENAME,
)
from .util import (
    FileOperator,
    convert_path,
    CSVReader,
    CSVWriter,
    Cache,
    cached_property,
    get_cache_base,
    read_exports,
    tempdir,
    get_platform,
)
from .version import NormalizedVersion, UnsupportedVersionError

logger = logging.getLogger(__name__)

cache = None  # created when needed

if hasattr(sys, "pypy_version_info"):  # pragma: no cover
    IMP_PREFIX = "pp"
elif sys.platform.startswith("java"):  # pragma: no cover
    IMP_PREFIX = "jy"
elif sys.platform == "cli":  # pragma: no cover
    IMP_PREFIX = "ip"
else:
    IMP_PREFIX = "cp"

VER_SUFFIX = sysconfig.get_config_var("py_version_nodot")
if not VER_SUFFIX:  # pragma: no cover
    VER_SUFFIX = "%s%s" % sys.version_info[:2]
PYVER = "py" + VER_SUFFIX
IMPVER = IMP_PREFIX + VER_SUFFIX

ARCH = get_platform().replace("-", "_").replace(".", "_")

ABI = sysconfig.get_config_var("SOABI")
if ABI and ABI.startswith("cpython-"):
    ABI = ABI.replace("cpython-", "cp").split("-")[0]
else:

    def _derive_abi():
        parts = ["cp", VER_SUFFIX]
        if sysconfig.get_config_var("Py_DEBUG"):
            parts.append("d")
        if IMP_PREFIX == "cp":
            vi = sys.version_info[:2]
            if vi < (3, 8):
                wpm = sysconfig.get_config_var("WITH_PYMALLOC")
                if wpm is None:
                    wpm = True
                if wpm:
                    parts.append("m")
                if vi < (3, 3):
                    us = sysconfig.get_config_var("Py_UNICODE_SIZE")
                    if us == 4 or (us is None and sys.maxunicode == 0x10FFFF):
                        parts.append("u")
        return "".join(parts)

    ABI = _derive_abi()
    del _derive_abi

FILENAME_RE = re.compile(
    r"""
(?P<nm>[^-]+)
-(?P<vn>\d+[^-]*)
(-(?P<bn>\d+[^-]*))?
-(?P<py>\w+\d+(\.\w+\d+)*)
-(?P<bi>\w+)
-(?P<ar>\w+(\.\w+)*)
\.whl$
""",
    re.IGNORECASE | re.VERBOSE,
)

NAME_VERSION_RE = re.compile(
    r"""
(?P<nm>[^-]+)
-(?P<vn>\d+[^-]*)
(-(?P<bn>\d+[^-]*))?$
""",
    re.IGNORECASE | re.VERBOSE,
)

SHEBANG_RE = re.compile(rb"\s*#![^\r\n]*")
SHEBANG_DETAIL_RE = re.compile(rb'^(\s*#!("[^"]+"|\S+))\s+(.*)$')
SHEBANG_PYTHON = b"#!python"
SHEBANG_PYTHONW = b"#!pythonw"

if os.sep == "/":
    to_posix = lambda o: o
else:
    to_posix = lambda o: o.replace(os.sep, "/")

if sys.version_info[0] < 3:
    import imp
else:
    imp = None
    import importlib.machinery
    import importlib.util


def _get_suffixes():
    if imp:
        return [s[0] for s in imp.get_suffixes()]
    else:
        return importlib.machinery.EXTENSION_SUFFIXES


def _load_dynamic(name, path):
    # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    if imp:
        return imp.load_dynamic(name, path)
    else:
        spec = importlib.util.spec_from_file_location(name, path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[name] = module
        spec.loader.exec_module(module)
        return module


class Mounter(object):
    def __init__(self):
        self.impure_wheels = {}
        self.libs = {}

    def add(self, pathname, extensions):
        self.impure_wheels[pathname] = extensions
        self.libs.update(extensions)

    def remove(self, pathname):
        extensions = self.impure_wheels.pop(pathname)
        for k, v in extensions:
            if k in self.libs:
                del self.libs[k]

    def find_module(self, fullname, path=None):
        if fullname in self.libs:
            result = self
        else:
            result = None
        return result

    def load_module(self, fullname):
        if fullname in sys.modules:
            result = sys.modules[fullname]
        else:
            if fullname not in self.libs:
                raise ImportError("unable to find extension for %s" % fullname)
            result = _load_dynamic(fullname, self.libs[fullname])
            result.__loader__ = self
            parts = fullname.rsplit(".", 1)
            if len(parts) > 1:
                result.__package__ = parts[0]
        return result


_hook = Mounter()


class Wheel(object):
    """
    Class to build and install from Wheel files (PEP 427).
    """

    wheel_version = (1, 1)
    hash_kind = "sha256"

    def __init__(self, filename=None, sign=False, verify=False):
        """
        Initialise an instance using a (valid) filename.
        """
        self.sign = sign
        self.should_verify = verify
        self.buildver = ""
        self.pyver = [PYVER]
        self.abi = ["none"]
        self.arch = ["any"]
        self.dirname = os.getcwd()
        if filename is None:
            self.name = "dummy"
            self.version = "0.1"
            self._filename = self.filename
        else:
            m = NAME_VERSION_RE.match(filename)
            if m:
                info = m.groupdict("")
                self.name = info["nm"]
                # Reinstate the local version separator
                self.version = info["vn"].replace("_", "-")
                self.buildver = info["bn"]
                self._filename = self.filename
            else:
                dirname, filename = os.path.split(filename)
                m = FILENAME_RE.match(filename)
                if not m:
                    raise DistlibException("Invalid name or " "filename: %r" % filename)
                if dirname:
                    self.dirname = os.path.abspath(dirname)
                self._filename = filename
                info = m.groupdict("")
                self.name = info["nm"]
                self.version = info["vn"]
                self.buildver = info["bn"]
                self.pyver = info["py"].split(".")
                self.abi = info["bi"].split(".")
                self.arch = info["ar"].split(".")

    @property
    def filename(self):
        """
        Build and return a filename from the various components.
        """
        if self.buildver:
            buildver = "-" + self.buildver
        else:
            buildver = ""
        pyver = ".".join(self.pyver)
        abi = ".".join(self.abi)
        arch = ".".join(self.arch)
        # replace - with _ as a local version separator
        version = self.version.replace("-", "_")
        return "%s-%s%s-%s-%s-%s.whl" % (self.name, version, buildver, pyver, abi, arch)

    @property
    def exists(self):
        path = os.path.join(self.dirname, self.filename)
        return os.path.isfile(path)

    @property
    def tags(self):
        for pyver in self.pyver:
            for abi in self.abi:
                for arch in self.arch:
                    yield pyver, abi, arch

    @cached_property
    def metadata(self):
        pathname = os.path.join(self.dirname, self.filename)
        name_ver = "%s-%s" % (self.name, self.version)
        info_dir = "%s.dist-info" % name_ver
        wrapper = codecs.getreader("utf-8")
        with ZipFile(pathname, "r") as zf:
            wheel_metadata = self.get_wheel_metadata(zf)
            wv = wheel_metadata["Wheel-Version"].split(".", 1)
            file_version = tuple([int(i) for i in wv])
            # if file_version < (1, 1):
            # fns = [WHEEL_METADATA_FILENAME, METADATA_FILENAME,
            # LEGACY_METADATA_FILENAME]
            # else:
            # fns = [WHEEL_METADATA_FILENAME, METADATA_FILENAME]
            fns = [WHEEL_METADATA_FILENAME, LEGACY_METADATA_FILENAME]
            result = None
            for fn in fns:
                try:
                    metadata_filename = posixpath.join(info_dir, fn)
                    with zf.open(metadata_filename) as bf:
                        wf = wrapper(bf)
                        result = Metadata(fileobj=wf)
                        if result:
                            break
                except KeyError:
                    pass
            if not result:
                raise ValueError(
                    "Invalid wheel, because metadata is "
                    "missing: looked in %s" % ", ".join(fns)
                )
        return result

    def get_wheel_metadata(self, zf):
        name_ver = "%s-%s" % (self.name, self.version)
        info_dir = "%s.dist-info" % name_ver
        metadata_filename = posixpath.join(info_dir, "WHEEL")
        with zf.open(metadata_filename) as bf:
            wf = codecs.getreader("utf-8")(bf)
            message = message_from_file(wf)
        return dict(message)

    @cached_property
    def info(self):
        pathname = os.path.join(self.dirname, self.filename)
        with ZipFile(pathname, "r") as zf:
            result = self.get_wheel_metadata(zf)
        return result

    def process_shebang(self, data):
        m = SHEBANG_RE.match(data)
        if m:
            end = m.end()
            shebang, data_after_shebang = data[:end], data[end:]
            # Preserve any arguments after the interpreter
            if b"pythonw" in shebang.lower():
                shebang_python = SHEBANG_PYTHONW
            else:
                shebang_python = SHEBANG_PYTHON
            m = SHEBANG_DETAIL_RE.match(shebang)
            if m:
                args = b" " + m.groups()[-1]
            else:
                args = b""
            shebang = shebang_python + args
            data = shebang + data_after_shebang
        else:
            cr = data.find(b"\r")
            lf = data.find(b"\n")
            if cr < 0 or cr > lf:
                term = b"\n"
            else:
                if data[cr : cr + 2] == b"\r\n":
                    term = b"\r\n"
                else:
                    term = b"\r"
            data = SHEBANG_PYTHON + term + data
        return data

    def get_hash(self, data, hash_kind=None):
        if hash_kind is None:
            hash_kind = self.hash_kind
        try:
            hasher = getattr(hashlib, hash_kind)
        except AttributeError:
            raise DistlibException("Unsupported hash algorithm: %r" % hash_kind)
        result = hasher(data).digest()
        result = base64.urlsafe_b64encode(result).rstrip(b"=").decode("ascii")
        return hash_kind, result

    def write_record(self, records, record_path, archive_record_path):
        records = list(records)  # make a copy, as mutated
        records.append((archive_record_path, "", ""))
        with CSVWriter(record_path) as writer:
            for row in records:
                writer.writerow(row)

    def write_records(self, info, libdir, archive_paths):
        records = []
        distinfo, info_dir = info
        hasher = getattr(hashlib, self.hash_kind)
        for ap, p in archive_paths:
            with open(p, "rb") as f:
                data = f.read()
            digest = "%s=%s" % self.get_hash(data)
            size = os.path.getsize(p)
            records.append((ap, digest, size))

        p = os.path.join(distinfo, "RECORD")
        ap = to_posix(os.path.join(info_dir, "RECORD"))
        self.write_record(records, p, ap)
        archive_paths.append((ap, p))

    def build_zip(self, pathname, archive_paths):
        with ZipFile(pathname, "w", zipfile.ZIP_DEFLATED) as zf:
            for ap, p in archive_paths:
                logger.debug("Wrote %s to %s in wheel", p, ap)
                zf.write(p, ap)

    def build(self, paths, tags=None, wheel_version=None):
        """
        Build a wheel from files in specified paths, and use any specified tags
        when determining the name of the wheel.
        """
        if tags is None:
            tags = {}

        libkey = list(filter(lambda o: o in paths, ("purelib", "platlib")))[0]
        if libkey == "platlib":
            is_pure = "false"
            default_pyver = [IMPVER]
            default_abi = [ABI]
            default_arch = [ARCH]
        else:
            is_pure = "true"
            default_pyver = [PYVER]
            default_abi = ["none"]
            default_arch = ["any"]

        self.pyver = tags.get("pyver", default_pyver)
        self.abi = tags.get("abi", default_abi)
        self.arch = tags.get("arch", default_arch)

        libdir = paths[libkey]

        name_ver = "%s-%s" % (self.name, self.version)
        data_dir = "%s.data" % name_ver
        info_dir = "%s.dist-info" % name_ver

        archive_paths = []

        # First, stuff which is not in site-packages
        for key in ("data", "headers", "scripts"):
            if key not in paths:
                continue
            path = paths[key]
            if os.path.isdir(path):
                for root, dirs, files in os.walk(path):
                    for fn in files:
                        p = fsdecode(os.path.join(root, fn))
                        rp = os.path.relpath(p, path)
                        ap = to_posix(os.path.join(data_dir, key, rp))
                        archive_paths.append((ap, p))
                        if key == "scripts" and not p.endswith(".exe"):
                            with open(p, "rb") as f:
                                data = f.read()
                            data = self.process_shebang(data)
                            with open(p, "wb") as f:
                                f.write(data)

        # Now, stuff which is in site-packages, other than the
        # distinfo stuff.
        path = libdir
        distinfo = None
        for root, dirs, files in os.walk(path):
            if root == path:
                # At the top level only, save distinfo for later
                # and skip it for now
                for i, dn in enumerate(dirs):
                    dn = fsdecode(dn)
                    if dn.endswith(".dist-info"):
                        distinfo = os.path.join(root, dn)
                        del dirs[i]
                        break
                assert distinfo, ".dist-info directory expected, not found"

            for fn in files:
                # comment out next suite to leave .pyc files in
                if fsdecode(fn).endswith((".pyc", ".pyo")):
                    continue
                p = os.path.join(root, fn)
                rp = to_posix(os.path.relpath(p, path))
                archive_paths.append((rp, p))

        # Now distinfo. Assumed to be flat, i.e. os.listdir is enough.
        files = os.listdir(distinfo)
        for fn in files:
            if fn not in ("RECORD", "INSTALLER", "SHARED", "WHEEL"):
                p = fsdecode(os.path.join(distinfo, fn))
                ap = to_posix(os.path.join(info_dir, fn))
                archive_paths.append((ap, p))

        wheel_metadata = [
            "Wheel-Version: %d.%d" % (wheel_version or self.wheel_version),
            "Generator: distlib %s" % __version__,
            "Root-Is-Purelib: %s" % is_pure,
        ]
        for pyver, abi, arch in self.tags:
            wheel_metadata.append("Tag: %s-%s-%s" % (pyver, abi, arch))
        p = os.path.join(distinfo, "WHEEL")
        with open(p, "w") as f:
            f.write("\n".join(wheel_metadata))
        ap = to_posix(os.path.join(info_dir, "WHEEL"))
        archive_paths.append((ap, p))

        # sort the entries by archive path. Not needed by any spec, but it
        # keeps the archive listing and RECORD tidier than they would otherwise
        # be. Use the number of path segments to keep directory entries together,
        # and keep the dist-info stuff at the end.
        def sorter(t):
            ap = t[0]
            n = ap.count("/")
            if ".dist-info" in ap:
                n += 10000
            return (n, ap)

        archive_paths = sorted(archive_paths, key=sorter)

        # Now, at last, RECORD.
        # Paths in here are archive paths - nothing else makes sense.
        self.write_records((distinfo, info_dir), libdir, archive_paths)
        # Now, ready to build the zip file
        pathname = os.path.join(self.dirname, self.filename)
        self.build_zip(pathname, archive_paths)
        return pathname

    def skip_entry(self, arcname):
        """
        Determine whether an archive entry should be skipped when verifying
        or installing.
        """
        # The signature file won't be in RECORD,
        # and we  don't currently don't do anything with it
        # We also skip directories, as they won't be in RECORD
        # either. See:
        #
        # https://github.com/pypa/wheel/issues/294
        # https://github.com/pypa/wheel/issues/287
        # https://github.com/pypa/wheel/pull/289
        #
        return arcname.endswith(("/", "/RECORD.jws"))

    def install(self, paths, maker, **kwargs):
        """
        Install a wheel to the specified paths. If kwarg ``warner`` is
        specified, it should be a callable, which will be called with two
        tuples indicating the wheel version of this software and the wheel
        version in the file, if there is a discrepancy in the versions.
        This can be used to issue any warnings to raise any exceptions.
        If kwarg ``lib_only`` is True, only the purelib/platlib files are
        installed, and the headers, scripts, data and dist-info metadata are
        not written. If kwarg ``bytecode_hashed_invalidation`` is True, written
        bytecode will try to use file-hash based invalidation (PEP-552) on
        supported interpreter versions (CPython 2.7+).

        The return value is a :class:`InstalledDistribution` instance unless
        ``options.lib_only`` is True, in which case the return value is ``None``.
        """

        dry_run = maker.dry_run
        warner = kwargs.get("warner")
        lib_only = kwargs.get("lib_only", False)
        bc_hashed_invalidation = kwargs.get("bytecode_hashed_invalidation", False)

        pathname = os.path.join(self.dirname, self.filename)
        name_ver = "%s-%s" % (self.name, self.version)
        data_dir = "%s.data" % name_ver
        info_dir = "%s.dist-info" % name_ver

        metadata_name = posixpath.join(info_dir, LEGACY_METADATA_FILENAME)
        wheel_metadata_name = posixpath.join(info_dir, "WHEEL")
        record_name = posixpath.join(info_dir, "RECORD")

        wrapper = codecs.getreader("utf-8")

        with ZipFile(pathname, "r") as zf:
            with zf.open(wheel_metadata_name) as bwf:
                wf = wrapper(bwf)
                message = message_from_file(wf)
            wv = message["Wheel-Version"].split(".", 1)
            file_version = tuple([int(i) for i in wv])
            if (file_version != self.wheel_version) and warner:
                warner(self.wheel_version, file_version)

            if message["Root-Is-Purelib"] == "true":
                libdir = paths["purelib"]
            else:
                libdir = paths["platlib"]

            records = {}
            with zf.open(record_name) as bf:
                with CSVReader(stream=bf) as reader:
                    for row in reader:
                        p = row[0]
                        records[p] = row

            data_pfx = posixpath.join(data_dir, "")
            info_pfx = posixpath.join(info_dir, "")
            script_pfx = posixpath.join(data_dir, "scripts", "")

            # make a new instance rather than a copy of maker's,
            # as we mutate it
            fileop = FileOperator(dry_run=dry_run)
            fileop.record = True  # so we can rollback if needed

            bc = not sys.dont_write_bytecode  # Double negatives. Lovely!

            outfiles = []  # for RECORD writing

            # for script copying/shebang processing
            workdir = tempfile.mkdtemp()
            # set target dir later
            # we default add_launchers to False, as the
            # Python Launcher should be used instead
            maker.source_dir = workdir
            maker.target_dir = None
            try:
                for zinfo in zf.infolist():
                    arcname = zinfo.filename
                    if isinstance(arcname, text_type):
                        u_arcname = arcname
                    else:
                        u_arcname = arcname.decode("utf-8")
                    if self.skip_entry(u_arcname):
                        continue
                    row = records[u_arcname]
                    if row[2] and str(zinfo.file_size) != row[2]:
                        raise DistlibException("size mismatch for " "%s" % u_arcname)
                    if row[1]:
                        kind, value = row[1].split("=", 1)
                        with zf.open(arcname) as bf:
                            data = bf.read()
                        _, digest = self.get_hash(data, kind)
                        if digest != value:
                            raise DistlibException(
                                "digest mismatch for " "%s" % arcname
                            )

                    if lib_only and u_arcname.startswith((info_pfx, data_pfx)):
                        logger.debug("lib_only: skipping %s", u_arcname)
                        continue
                    is_script = u_arcname.startswith(
                        script_pfx
                    ) and not u_arcname.endswith(".exe")

                    if u_arcname.startswith(data_pfx):
                        _, where, rp = u_arcname.split("/", 2)
                        outfile = os.path.join(paths[where], convert_path(rp))
                    else:
                        # meant for site-packages.
                        if u_arcname in (wheel_metadata_name, record_name):
                            continue
                        outfile = os.path.join(libdir, convert_path(u_arcname))
                    if not is_script:
                        with zf.open(arcname) as bf:
                            fileop.copy_stream(bf, outfile)
                        # Issue #147: permission bits aren't preserved. Using
                        # zf.extract(zinfo, libdir) should have worked, but didn't,
                        # see https://www.thetopsites.net/article/53834422.shtml
                        # So ... manually preserve permission bits as given in zinfo
                        if os.name == "posix":
                            # just set the normal permission bits
                            os.chmod(outfile, (zinfo.external_attr >> 16) & 0x1FF)
                        outfiles.append(outfile)
                        # Double check the digest of the written file
                        if not dry_run and row[1]:
                            with open(outfile, "rb") as bf:
                                data = bf.read()
                                _, newdigest = self.get_hash(data, kind)
                                if newdigest != digest:
                                    raise DistlibException(
                                        "digest mismatch "
                                        "on write for "
                                        "%s" % outfile
                                    )
                        if bc and outfile.endswith(".py"):
                            try:
                                pyc = fileop.byte_compile(
                                    outfile, hashed_invalidation=bc_hashed_invalidation
                                )
                                outfiles.append(pyc)
                            except Exception:
                                # Don't give up if byte-compilation fails,
                                # but log it and perhaps warn the user
                                logger.warning("Byte-compilation failed", exc_info=True)
                    else:
                        fn = os.path.basename(convert_path(arcname))
                        workname = os.path.join(workdir, fn)
                        with zf.open(arcname) as bf:
                            fileop.copy_stream(bf, workname)

                        dn, fn = os.path.split(outfile)
                        maker.target_dir = dn
                        filenames = maker.make(fn)
                        fileop.set_executable_mode(filenames)
                        outfiles.extend(filenames)

                if lib_only:
                    logger.debug("lib_only: returning None")
                    dist = None
                else:
                    # Generate scripts

                    # Try to get pydist.json so we can see if there are
                    # any commands to generate. If this fails (e.g. because
                    # of a legacy wheel), log a warning but don't give up.
                    commands = None
                    file_version = self.info["Wheel-Version"]
                    if file_version == "1.0":
                        # Use legacy info
                        ep = posixpath.join(info_dir, "entry_points.txt")
                        try:
                            with zf.open(ep) as bwf:
                                epdata = read_exports(bwf)
                            commands = {}
                            for key in ("console", "gui"):
                                k = "%s_scripts" % key
                                if k in epdata:
                                    commands["wrap_%s" % key] = d = {}
                                    for v in epdata[k].values():
                                        s = "%s:%s" % (v.prefix, v.suffix)
                                        if v.flags:
                                            s += " [%s]" % ",".join(v.flags)
                                        d[v.name] = s
                        except Exception:
                            logger.warning(
                                "Unable to read legacy script "
                                "metadata, so cannot generate "
                                "scripts"
                            )
                    else:
                        try:
                            with zf.open(metadata_name) as bwf:
                                wf = wrapper(bwf)
                                commands = json.load(wf).get("extensions")
                                if commands:
                                    commands = commands.get("python.commands")
                        except Exception:
                            logger.warning(
                                "Unable to read JSON metadata, so "
                                "cannot generate scripts"
                            )
                    if commands:
                        console_scripts = commands.get("wrap_console", {})
                        gui_scripts = commands.get("wrap_gui", {})
                        if console_scripts or gui_scripts:
                            script_dir = paths.get("scripts", "")
                            if not os.path.isdir(script_dir):
                                raise ValueError("Valid script path not " "specified")
                            maker.target_dir = script_dir
                            for k, v in console_scripts.items():
                                script = "%s = %s" % (k, v)
                                filenames = maker.make(script)
                                fileop.set_executable_mode(filenames)

                            if gui_scripts:
                                options = {"gui": True}
                                for k, v in gui_scripts.items():
                                    script = "%s = %s" % (k, v)
                                    filenames = maker.make(script, options)
                                    fileop.set_executable_mode(filenames)

                    p = os.path.join(libdir, info_dir)
                    dist = InstalledDistribution(p)

                    # Write SHARED
                    paths = dict(paths)  # don't change passed in dict
                    del paths["purelib"]
                    del paths["platlib"]
                    paths["lib"] = libdir
                    p = dist.write_shared_locations(paths, dry_run)
                    if p:
                        outfiles.append(p)

                    # Write RECORD
                    dist.write_installed_files(outfiles, paths["prefix"], dry_run)
                return dist
            except Exception:  # pragma: no cover
                logger.exception("installation failed.")
                fileop.rollback()
                raise
            finally:
                shutil.rmtree(workdir)

    def _get_dylib_cache(self):
        global cache
        if cache is None:
            # Use native string to avoid issues on 2.x: see Python #20140.
            base = os.path.join(
                get_cache_base(), str("dylib-cache"), "%s.%s" % sys.version_info[:2]
            )
            cache = Cache(base)
        return cache

    def _get_extensions(self):
        pathname = os.path.join(self.dirname, self.filename)
        name_ver = "%s-%s" % (self.name, self.version)
        info_dir = "%s.dist-info" % name_ver
        arcname = posixpath.join(info_dir, "EXTENSIONS")
        wrapper = codecs.getreader("utf-8")
        result = []
        with ZipFile(pathname, "r") as zf:
            try:
                with zf.open(arcname) as bf:
                    wf = wrapper(bf)
                    extensions = json.load(wf)
                    cache = self._get_dylib_cache()
                    prefix = cache.prefix_to_dir(pathname)
                    cache_base = os.path.join(cache.base, prefix)
                    if not os.path.isdir(cache_base):
                        os.makedirs(cache_base)
                    for name, relpath in extensions.items():
                        dest = os.path.join(cache_base, convert_path(relpath))
                        if not os.path.exists(dest):
                            extract = True
                        else:
                            file_time = os.stat(dest).st_mtime
                            file_time = datetime.datetime.fromtimestamp(file_time)
                            info = zf.getinfo(relpath)
                            wheel_time = datetime.datetime(*info.date_time)
                            extract = wheel_time > file_time
                        if extract:
                            zf.extract(relpath, cache_base)
                        result.append((name, dest))
            except KeyError:
                pass
        return result

    def is_compatible(self):
        """
        Determine if a wheel is compatible with the running system.
        """
        return is_compatible(self)

    def is_mountable(self):
        """
        Determine if a wheel is asserted as mountable by its metadata.
        """
        return True  # for now - metadata details TBD

    def mount(self, append=False):
        pathname = os.path.abspath(os.path.join(self.dirname, self.filename))
        if not self.is_compatible():
            msg = "Wheel %s not compatible with this Python." % pathname
            raise DistlibException(msg)
        if not self.is_mountable():
            msg = "Wheel %s is marked as not mountable." % pathname
            raise DistlibException(msg)
        if pathname in sys.path:
            logger.debug("%s already in path", pathname)
        else:
            if append:
                sys.path.append(pathname)
            else:
                sys.path.insert(0, pathname)
            extensions = self._get_extensions()
            if extensions:
                if _hook not in sys.meta_path:
                    sys.meta_path.append(_hook)
                _hook.add(pathname, extensions)

    def unmount(self):
        pathname = os.path.abspath(os.path.join(self.dirname, self.filename))
        if pathname not in sys.path:
            logger.debug("%s not in path", pathname)
        else:
            sys.path.remove(pathname)
            if pathname in _hook.impure_wheels:
                _hook.remove(pathname)
            if not _hook.impure_wheels:
                if _hook in sys.meta_path:
                    sys.meta_path.remove(_hook)

    def verify(self):
        pathname = os.path.join(self.dirname, self.filename)
        name_ver = "%s-%s" % (self.name, self.version)
        data_dir = "%s.data" % name_ver
        info_dir = "%s.dist-info" % name_ver

        metadata_name = posixpath.join(info_dir, LEGACY_METADATA_FILENAME)
        wheel_metadata_name = posixpath.join(info_dir, "WHEEL")
        record_name = posixpath.join(info_dir, "RECORD")

        wrapper = codecs.getreader("utf-8")

        with ZipFile(pathname, "r") as zf:
            with zf.open(wheel_metadata_name) as bwf:
                wf = wrapper(bwf)
                message = message_from_file(wf)
            wv = message["Wheel-Version"].split(".", 1)
            file_version = tuple([int(i) for i in wv])
            # TODO version verification

            records = {}
            with zf.open(record_name) as bf:
                with CSVReader(stream=bf) as reader:
                    for row in reader:
                        p = row[0]
                        records[p] = row

            for zinfo in zf.infolist():
                arcname = zinfo.filename
                if isinstance(arcname, text_type):
                    u_arcname = arcname
                else:
                    u_arcname = arcname.decode("utf-8")
                # See issue #115: some wheels have .. in their entries, but
                # in the filename ... e.g. __main__..py ! So the check is
                # updated to look for .. in the directory portions
                p = u_arcname.split("/")
                if ".." in p:
                    raise DistlibException("invalid entry in " "wheel: %r" % u_arcname)

                if self.skip_entry(u_arcname):
                    continue
                row = records[u_arcname]
                if row[2] and str(zinfo.file_size) != row[2]:
                    raise DistlibException("size mismatch for " "%s" % u_arcname)
                if row[1]:
                    kind, value = row[1].split("=", 1)
                    with zf.open(arcname) as bf:
                        data = bf.read()
                    _, digest = self.get_hash(data, kind)
                    if digest != value:
                        raise DistlibException("digest mismatch for " "%s" % arcname)

    def update(self, modifier, dest_dir=None, **kwargs):
        """
        Update the contents of a wheel in a generic way. The modifier should
        be a callable which expects a dictionary argument: its keys are
        archive-entry paths, and its values are absolute filesystem paths
        where the contents the corresponding archive entries can be found. The
        modifier is free to change the contents of the files pointed to, add
        new entries and remove entries, before returning. This method will
        extract the entire contents of the wheel to a temporary location, call
        the modifier, and then use the passed (and possibly updated)
        dictionary to write a new wheel. If ``dest_dir`` is specified, the new
        wheel is written there -- otherwise, the original wheel is overwritten.

        The modifier should return True if it updated the wheel, else False.
        This method returns the same value the modifier returns.
        """

        def get_version(path_map, info_dir):
            version = path = None
            key = "%s/%s" % (info_dir, LEGACY_METADATA_FILENAME)
            if key not in path_map:
                key = "%s/PKG-INFO" % info_dir
            if key in path_map:
                path = path_map[key]
                version = Metadata(path=path).version
            return version, path

        def update_version(version, path):
            updated = None
            try:
                v = NormalizedVersion(version)
                i = version.find("-")
                if i < 0:
                    updated = "%s+1" % version
                else:
                    parts = [int(s) for s in version[i + 1 :].split(".")]
                    parts[-1] += 1
                    updated = "%s+%s" % (version[:i], ".".join(str(i) for i in parts))
            except UnsupportedVersionError:
                logger.debug(
                    "Cannot update non-compliant (PEP-440) " "version %r", version
                )
            if updated:
                md = Metadata(path=path)
                md.version = updated
                legacy = path.endswith(LEGACY_METADATA_FILENAME)
                md.write(path=path, legacy=legacy)
                logger.debug("Version updated from %r to %r", version, updated)

        pathname = os.path.join(self.dirname, self.filename)
        name_ver = "%s-%s" % (self.name, self.version)
        info_dir = "%s.dist-info" % name_ver
        record_name = posixpath.join(info_dir, "RECORD")
        with tempdir() as workdir:
            with ZipFile(pathname, "r") as zf:
                path_map = {}
                for zinfo in zf.infolist():
                    arcname = zinfo.filename
                    if isinstance(arcname, text_type):
                        u_arcname = arcname
                    else:
                        u_arcname = arcname.decode("utf-8")
                    if u_arcname == record_name:
                        continue
                    if ".." in u_arcname:
                        raise DistlibException(
                            "invalid entry in " "wheel: %r" % u_arcname
                        )
                    zf.extract(zinfo, workdir)
                    path = os.path.join(workdir, convert_path(u_arcname))
                    path_map[u_arcname] = path

            # Remember the version.
            original_version, _ = get_version(path_map, info_dir)
            # Files extracted. Call the modifier.
            modified = modifier(path_map, **kwargs)
            if modified:
                # Something changed - need to build a new wheel.
                current_version, path = get_version(path_map, info_dir)
                if current_version and (current_version == original_version):
                    # Add or update local version to signify changes.
                    update_version(current_version, path)
                # Decide where the new wheel goes.
                if dest_dir is None:
                    fd, newpath = tempfile.mkstemp(
                        suffix=".whl", prefix="wheel-update-", dir=workdir
                    )
                    os.close(fd)
                else:
                    if not os.path.isdir(dest_dir):
                        raise DistlibException("Not a directory: %r" % dest_dir)
                    newpath = os.path.join(dest_dir, self.filename)
                archive_paths = list(path_map.items())
                distinfo = os.path.join(workdir, info_dir)
                info = distinfo, info_dir
                self.write_records(info, workdir, archive_paths)
                self.build_zip(newpath, archive_paths)
                if dest_dir is None:
                    shutil.copyfile(newpath, pathname)
        return modified


def _get_glibc_version():
    import platform

    ver = platform.libc_ver()
    result = []
    if ver[0] == "glibc":
        for s in ver[1].split("."):
            result.append(int(s) if s.isdigit() else 0)
        result = tuple(result)
    return result


def compatible_tags():
    """
    Return (pyver, abi, arch) tuples compatible with this Python.
    """
    versions = [VER_SUFFIX]
    major = VER_SUFFIX[0]
    for minor in range(sys.version_info[1] - 1, -1, -1):
        versions.append("".join([major, str(minor)]))

    abis = []
    for suffix in _get_suffixes():
        if suffix.startswith(".abi"):
            abis.append(suffix.split(".", 2)[1])
    abis.sort()
    if ABI != "none":
        abis.insert(0, ABI)
    abis.append("none")
    result = []

    arches = [ARCH]
    if sys.platform == "darwin":
        m = re.match(r"(\w+)_(\d+)_(\d+)_(\w+)$", ARCH)
        if m:
            name, major, minor, arch = m.groups()
            minor = int(minor)
            matches = [arch]
            if arch in ("i386", "ppc"):
                matches.append("fat")
            if arch in ("i386", "ppc", "x86_64"):
                matches.append("fat3")
            if arch in ("ppc64", "x86_64"):
                matches.append("fat64")
            if arch in ("i386", "x86_64"):
                matches.append("intel")
            if arch in ("i386", "x86_64", "intel", "ppc", "ppc64"):
                matches.append("universal")
            while minor >= 0:
                for match in matches:
                    s = "%s_%s_%s_%s" % (name, major, minor, match)
                    if s != ARCH:  # already there
                        arches.append(s)
                minor -= 1

    # Most specific - our Python version, ABI and arch
    for abi in abis:
        for arch in arches:
            result.append(("".join((IMP_PREFIX, versions[0])), abi, arch))
            # manylinux
            if abi != "none" and sys.platform.startswith("linux"):
                arch = arch.replace("linux_", "")
                parts = _get_glibc_version()
                if len(parts) == 2:
                    if parts >= (2, 5):
                        result.append(
                            (
                                "".join((IMP_PREFIX, versions[0])),
                                abi,
                                "manylinux1_%s" % arch,
                            )
                        )
                    if parts >= (2, 12):
                        result.append(
                            (
                                "".join((IMP_PREFIX, versions[0])),
                                abi,
                                "manylinux2010_%s" % arch,
                            )
                        )
                    if parts >= (2, 17):
                        result.append(
                            (
                                "".join((IMP_PREFIX, versions[0])),
                                abi,
                                "manylinux2014_%s" % arch,
                            )
                        )
                    result.append(
                        (
                            "".join((IMP_PREFIX, versions[0])),
                            abi,
                            "manylinux_%s_%s_%s" % (parts[0], parts[1], arch),
                        )
                    )

    # where no ABI / arch dependency, but IMP_PREFIX dependency
    for i, version in enumerate(versions):
        result.append(("".join((IMP_PREFIX, version)), "none", "any"))
        if i == 0:
            result.append(("".join((IMP_PREFIX, version[0])), "none", "any"))

    # no IMP_PREFIX, ABI or arch dependency
    for i, version in enumerate(versions):
        result.append(("".join(("py", version)), "none", "any"))
        if i == 0:
            result.append(("".join(("py", version[0])), "none", "any"))

    return set(result)


COMPATIBLE_TAGS = compatible_tags()

del compatible_tags


def is_compatible(wheel, tags=None):
    if not isinstance(wheel, Wheel):
        wheel = Wheel(wheel)  # assume it's a filename
    result = False
    if tags is None:
        tags = COMPATIBLE_TAGS
    for ver, abi, arch in tags:
        if ver in wheel.pyver and abi in wheel.abi and arch in wheel.arch:
            result = True
            break
    return result



import io
import os
import zipfile


def dir_to_zipfile(root):
    """Construct an in-memory zip file for a directory."""
    buffer = io.BytesIO()
    zip_file = zipfile.ZipFile(buffer, "w")
    for root, dirs, files in os.walk(root):
        for path in dirs:
            fs_path = os.path.join(root, path)
            rel_path = os.path.relpath(fs_path, root)
            zip_file.writestr(rel_path + "/", "")
        for path in files:
            fs_path = os.path.join(root, path)
            rel_path = os.path.relpath(fs_path, root)
            zip_file.write(fs_path, rel_path)
    return zip_file



# coding: utf-8
from .exceptions import *
from .ext import ExtType, Timestamp

import os
import sys


version = (1, 0, 4)
__version__ = "1.0.4"


if os.environ.get("MSGPACK_PUREPYTHON") or sys.version_info[0] == 2:
    from .fallback import Packer, unpackb, Unpacker
else:
    try:
        from ._cmsgpack import Packer, unpackb, Unpacker
    except ImportError:
        from .fallback import Packer, unpackb, Unpacker


def pack(o, stream, **kwargs):
    """
    Pack object `o` and write it to `stream`

    See :class:`Packer` for options.
    """
    packer = Packer(**kwargs)
    stream.write(packer.pack(o))


def packb(o, **kwargs):
    """
    Pack object `o` and return packed bytes

    See :class:`Packer` for options.
    """
    return Packer(**kwargs).pack(o)


def unpack(stream, **kwargs):
    """
    Unpack an object from `stream`.

    Raises `ExtraData` when `stream` contains extra bytes.
    See :class:`Unpacker` for options.
    """
    data = stream.read()
    return unpackb(data, **kwargs)


# alias for compatibility to simplejson/marshal/pickle.
load = unpack
loads = unpackb

dump = pack
dumps = packb
